
import jax.numpy as jnp
#from jax.scipy.special import gamma
from microjax.fastlens.special import gamma
from jax.numpy.fft import rfft, irfft
from jax import jit, vmap, lax
from functools import partial

class fftlog(object):
    def __init__(self, x, fx, nu=1.1, N_extrap_low=0, N_extrap_high=0, c_window_width=0.25, N_pad=0):
        self.x_origin = x # x is logarithmically spaced
        self.dlnx = jnp.log(x[1]/x[0])
        self.fx_origin= fx # f(x) array
        self.nu = nu
        self.N_extrap_low = N_extrap_low
        self.N_extrap_high = N_extrap_high
        self.c_window_width = c_window_width

        # extrapolate x and f(x) linearly in log(x), and log(f(x))
        self.x = log_extrap(x, N_extrap_low, N_extrap_high)
        self.fx = log_extrap(fx, N_extrap_low, N_extrap_high)
        self.N = self.x.size

        # zero-padding
        self.N_pad = N_pad
        if(N_pad):
            pad = jnp.zeros(N_pad)
            self.x = log_extrap(self.x, N_pad, N_pad)
            self.fx = jnp.hstack((pad, self.fx, pad))
            self.N += 2*N_pad
            self.N_extrap_high += N_pad
            self.N_extrap_low += N_pad
        
        if(self.N%2==1): # Make sure the array sizes are even
            self.x= self.x[:-1]
            self.fx=self.fx[:-1]
            self.N -= 1
            if(N_pad):
                self.N_extrap_high -=1 

        self.m, self.c_m = self.get_c_m()
        self.eta_m = 2*jnp.pi/(float(self.N)*self.dlnx) * self.m     
    #@jit
    def get_c_m(self):
        """
        return m and c_m
        c_m: the smoothed FFT coefficients of "biased" input function f(x): f_b = f(x) / x^\nu
        number of x values should be even
        c_window_width: the fraction of c_m elements that are smoothed,
        e.g. c_window_width=0.25 means smoothing the last 1/4 of c_m elements using "c_window".
        """
        f_b=self.fx * self.x**(-self.nu)
        c_m=rfft(f_b)
        m=jnp.arange(0,self.N//2+1) 
        c_m = c_m*c_window(m, int(self.c_window_width*self.N//2.) )
        return m, c_m
    #@jit 
    def fftlog(self, ell):
        """
        Calculate F(y) = \int_0^\infty dx / x * f(x) * j_\ell(xy),
        where j_\ell is the spherical Bessel func of order ell.
        array y is set as y[:] = (ell+1)/x[::-1]
        """
        z_ar = self.nu + 1j*self.eta_m
        y = (ell+1.) / self.x[::-1]
        h_m = self.c_m * (self.x[0]*y[0])**(-1j*self.eta_m) * g_l(ell, z_ar)

        Fy = irfft(jnp.conj(h_m)) * y**(-self.nu) * jnp.sqrt(jnp.pi)/4.
        #print(self.N_extrap_high,self.N,self.N_extrap_low)
        return y[self.N_extrap_high:self.N-self.N_extrap_low], Fy[self.N_extrap_high:self.N-self.N_extrap_low]

    def fftlog_dj(self, ell):
        """
        Calculate F(y) = \int_0^\infty dx / x * f(x) * j'_\ell(xy),
        where j_\ell is the spherical Bessel func of order ell.
        array y is set as y[:] = (ell+1)/x[::-1]
        """
        z_ar = self.nu + 1j*self.eta_m
        y = (ell+1.) / self.x[::-1]
        h_m = self.c_m * (self.x[0]*y[0])**(-1j*self.eta_m) * g_l_1(ell, z_ar)

        Fy = irfft(jnp.conj(h_m)) * y**(-self.nu) * jnp.sqrt(jnp.pi)/4.
        return y[self.N_extrap_high:self.N-self.N_extrap_low], Fy[self.N_extrap_high:self.N-self.N_extrap_low]
    
    def fftlog_ddj(self, ell):
        """
        Calculate F(y) = \int_0^\infty dx / x * f(x) * j''_\ell(xy),
        where j_\ell is the spherical Bessel func of order ell.
        array y is set as y[:] = (ell+1)/x[::-1]
        """
        z_ar = self.nu + 1j*self.eta_m
        y = (ell+1.) / self.x[::-1]
        h_m = self.c_m * (self.x[0]*y[0])**(-1j*self.eta_m) * g_l_2(ell, z_ar)

        Fy = irfft(jnp.conj(h_m)) * y**(-self.nu) * jnp.sqrt(jnp.pi)/4.
        return y[self.N_extrap_high:self.N-self.N_extrap_low], Fy[self.N_extrap_high:self.N-self.N_extrap_low]
    
    def fftlog_jsqr(self, ell):
        """
        Calculate F(y) = \int_0^\infty dx / x * f(x) * (j_\ell(xy))^2,
        where j_\ell is the spherical Bessel func of order ell.
        array y is set as y[:] = (ell+1)/x[::-1]
        """
        z_ar = self.nu + 1j*self.eta_m
        y = (ell+1.) / self.x[::-1]
        h_m = self.c_m * (self.x[0]*y[0])**(-1j*self.eta_m) * h_l(ell, z_ar)

        Fy = irfft(jnp.conj(h_m)) * y**(-self.nu) * jnp.sqrt(jnp.pi)/4.
        #print(self.N_extrap_high,self.N,self.N_extrap_low)
        return y[self.N_extrap_high:self.N-self.N_extrap_low], Fy[self.N_extrap_high:self.N-self.N_extrap_low]

class hankel(object):
    def __init__(self, x, fx, nu, N_extrap_low=0, N_extrap_high=0, c_window_width=0.25, N_pad=0):
        #print('nu is required to be between (0.5-n) and 2.')
        #print("HANKEL!!!",x)
        self.myfftlog = fftlog(x, jnp.sqrt(x)*fx, nu, N_extrap_low, N_extrap_high, c_window_width, N_pad)
    
    def hankel(self, n):
        y, Fy = self.myfftlog.fftlog(n-0.5)
        Fy *= jnp.sqrt(2*y/jnp.pi)
        return y, Fy

### Utility functions ####################
@partial(jit, static_argnums=(1,2))
def log_extrap(x, N_extrap_low, N_extrap_high):
    if N_extrap_low > 0:
        dlnx_low = jnp.log(x[1] / x[0])
        low_x = x[0] * jnp.exp(dlnx_low * jnp.arange(-N_extrap_low, 0))
    else:
        low_x = jnp.array([])

    if N_extrap_high > 0:
        dlnx_high = jnp.log(x[-1] / x[-2])
        high_x = x[-1] * jnp.exp(dlnx_high * jnp.arange(1, N_extrap_high + 1))
    else:
        high_x = jnp.array([])

    x_extrap = jnp.concatenate((low_x, x, high_x))
    return x_extrap    
"""
def log_extrap_(x, N_extrap_low, N_extrap_high):
    if x.size < 2:
         raise ValueError("x must have at least 2 elements")
    if x[0] == 0:
         raise ValueError("x[0] must be non-zero")
    if x[-2] == 0:
         raise ValueError("x[-2] must be non-zero")

    def compute(_):
        dlnx_low = jnp.log(x[1] / x[0])
        low_x = x[0] * jnp.exp(dlnx_low * jnp.arange(-N_extrap_low, 0))
        low_x = lax.cond(N_extrap_low > 0, lambda: low_x, lambda: jnp.zeros_like(low_x))
        dlnx_high = jnp.log(x[-1] / x[-2])
        high_x = x[-1] * jnp.exp(dlnx_high * jnp.arange(1, N_extrap_high + 1))
        high_x = lax.cond(N_extrap_high > 0, lambda: high_x, lambda: jnp.zeros_like(high_x))
        x_extrap = jnp.concatenate((low_x, x, high_x))
        return x_extrap
    def non_compute(_):
        return x
    return lax.cond((N_extrap_low > 0) | (N_extrap_high > 0), compute, non_compute, None)
    #return lax.cond(~jnp.logical_and(N_extrap_low==0, N_extrap_high==0), compute, non_compute, None)
"""

@partial(jit, static_argnums=(1,))
def c_window(n, n_cut):
    n_right = n[-1] - n_cut
    idx = n > n_right
    theta_right = (n[-1] - n) / (n[-1] - n_right - 1).astype(float)
    W = jnp.where(idx, theta_right - 1 / (2 * jnp.pi) * jnp.sin(2 * jnp.pi * theta_right), jnp.ones_like(n))
    return W

#@partial(jit, static_argnums=(0,))
@jit
def g_m_vals(mu, q):
    '''
    g_m_vals function adapted for JAX.
    g_m_vals(mu, q) = gamma( (mu+1+q)/2 ) / gamma( (mu+1-q)/2 ) = gamma(alpha+)/gamma(alpha-)
    mu = (alpha+) + (alpha-) - 1
    q = (alpha+) - (alpha-)

    Switching to asymptotic form when |Im(q)| + |mu| > cut = 200
    '''
    #if (mu + 1 + q.real[0] == 0):
    #    raise ValueError("gamma(0) encountered. Please change another nu value! Try nu=1.1.")

    imag_q = jnp.imag(q)
    g_m = jnp.zeros_like(q, dtype=complex)
    cut = 200
    mask_asym = jnp.abs(imag_q) + jnp.abs(mu) > cut
    mask_good = (jnp.abs(imag_q) + jnp.abs(mu) <= cut) & (q != mu + 1 + 0j)

    alpha_plus  = (mu + 1 + q) / 2.
    alpha_minus = (mu + 1 - q) / 2.
    g_m_good    = gamma(alpha_plus) / gamma(alpha_minus)

    asym_q = q
    asym_plus = (mu + 1 + asym_q) / 2.
    asym_minus = (mu + 1 - asym_q) / 2.
    g_m_asym = jnp.exp((asym_plus - 0.5) * jnp.log(asym_plus) - (asym_minus - 0.5) * jnp.log(asym_minus) - asym_q
                       + 1. / 12 * (1. / asym_plus - 1. / asym_minus) + 1. / 360 * (1. / asym_minus ** 3 - 1. / asym_plus ** 3)
                       + 1. / 1260 * (1. / asym_plus ** 5 - 1. / asym_minus ** 5))

    g_m = jnp.where(mask_good, g_m_good, g_m)
    g_m = jnp.where(mask_asym, g_m_asym, g_m)
    g_m = jnp.where(q == mu + 1 + 0j, jnp.zeros_like(g_m), g_m)

    return g_m

@jit
def g_m_ratio(a):
    '''
    g_m_ratio(a) = gamma(a)/gamma(a+0.5)
    switching to asymptotic form when |Im(a)| > cut = 200
    '''
    #if (a.real[0] == 0):
    #    raise ValueError("gamma(0) encountered. Please change another nu value! Try nu=1.1.")

    imag_a = jnp.imag(a)
    g_m = jnp.zeros_like(a, dtype=complex)
    cut = 100
    mask_asym = jnp.abs(imag_a) > cut
    mask_good = jnp.abs(imag_a) <= cut

    asym_a = a
    asym_a_plus = asym_a + 0.5
    g_m_asym = jnp.exp((asym_a - 0.5) * jnp.log(asym_a) - asym_a * jnp.log(asym_a_plus) + 0.5
                       + 1. / 12 * (1. / asym_a - 1. / asym_a_plus) + 1. / 360 * (1. / asym_a_plus ** 3 - 1. / asym_a ** 3)
                       + 1. / 1260 * (1. / asym_a ** 5 - 1. / asym_a_plus ** 5))

    a_good = a
    g_m_good = gamma(a_good) / gamma(a_good + 0.5)

    g_m = jnp.where(mask_good, g_m_good, g_m)
    g_m = jnp.where(mask_asym, g_m_asym, g_m)

    return g_m

def g_l(l,z_array):
    '''
	gl = 2^z_array * gamma((l+z_array)/2.) / gamma((3.+l-z_array)/2.)
	alpha+ = (l+z_array)/2.
	alpha- = (3.+l-z_array)/2.
	mu = (alpha+) + (alpha-) - 1 = l+0.5
	q = (alpha+) - (alpha-) = z_array - 1.5
	'''
    gl = 2.**z_array * g_m_vals(l+0.5,z_array-1.5)
    return gl

def g_l_1(l,z_array):
	'''
	for integral containing one first-derivative of spherical Bessel function
	gl1 = -2^(z_array-1) *(z_array -1)* gamma((l+z_array-1)/2.) / gamma((4.+l-z_array)/2.)
	mu = l+0.5
	q = z_array - 2.5
	'''
	gl1 = -2.**(z_array-1) *(z_array -1) * g_m_vals(l+0.5,z_array-2.5)
	return gl1

def g_l_2(l,z_array):
	'''
	for integral containing one 2nd-derivative of spherical Bessel function
	gl2 = 2^(z_array-2) *(z_array -1)*(z_array -2)* gamma((l+z_array-2)/2.) / gamma((5.+l-z_array)/2.)
	mu = l+0.5
	q = z_array - 3.5
	'''
	gl2 = 2.**(z_array-2) *(z_array -1)*(z_array -2)* g_m_vals(l+0.5,z_array-3.5)
	return gl2

def h_l(l,z_array):
	'''
	hl = gamma(l+ z_array/2.) * gamma((2.-z_array)/2.) / gamma((3.-z_array)/2.) / gamma(2.+l -z_array/2.)
	first component is g_m_vals(2l+1, z_array - 2)
	second component is gamma((2.-z_array)/2.) / gamma((3.-z_array)/2.)
	'''
	hl = g_m_vals(2*l+1., z_array - 2.) * g_m_ratio((2.-z_array)/2.)
	return hl


import numpy as np
import jax.numpy as jnp
from jax import jit, jacfwd, jacrev 
import jax
jax.config.update("jax_enable_x64", True)
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.ticker import AutoMinorLocator

from microjax.inverse_ray.extended_source import mag_uniform
from microjax.point_source import critical_and_caustic_curves

# Parameters
s  = 1.1  
q  = 0.1  
q3 = 0.03
r3_complex = 0.3 + 1.2j 
psi = jnp.arctan2(r3_complex.imag, r3_complex.real)

alpha = jnp.deg2rad(50) # angle between lens axis and source trajectory
tE = 10  # einstein radius crossing time
t0 = 0.0 # time of peak magnification
u0 = 0.1 # impact parameter
t  =  t0 + jnp.linspace(-0.5*tE, tE, 500)
rho = 0.02

tau = (t - t0)/tE
y1 = -u0*jnp.sin(alpha) + tau*jnp.cos(alpha)
y2 = u0*jnp.cos(alpha) + tau*jnp.sin(alpha)
w_points = jnp.array(y1 + 1j * y2, dtype=complex)

r_resolution  = 500
th_resolution = 500
Nlimb = 500
MAX_FULL_CALLS = 500
cubic = True

def chunked_vmap(func, data, chunk_size):
    results = []
    for i in range(0, len(data), chunk_size):
        chunk = data[i:i + chunk_size]
        results.append(jax.vmap(func)(chunk))
    return jnp.concatenate(results)

@jit
def get_mag(params):
    t0, tE, u0, q, s, alpha, rho, q3, r3, psi = params
    tau = (t - t0)/tE
    y1 = -u0*jnp.sin(alpha) + tau*jnp.cos(alpha)
    y2 = u0*jnp.cos(alpha) + tau*jnp.sin(alpha)
    w_points = jnp.array(y1 + 1j * y2, dtype=complex)
    _params = {"q": q, "s": s, "q3": q3, "r3": r3, "psi": psi}
    def mag_mj(w):
        return mag_uniform(w, rho, nlenses=3, **_params, cubic=cubic, 
                           r_resolution=r_resolution, th_resolution=th_resolution)
    magnifications = chunked_vmap(mag_mj, w_points, chunk_size=50)
    return w_points, magnifications

if(0):
    import time
    params = jnp.array([t0, tE, u0, q, s, alpha, rho, q3, jnp.abs(r3_complex), psi])
    get_mag(params)
    print("start")
    start = time.time()
    w_points, A = get_mag(params)
    end = time.time()
    print("mag finish: %.3f sec"%(end - start))
    # Evaluate the Jacobian at every point
    mag_jac = jit(jacfwd(lambda params: get_mag(params)[1]))
    mag_jac(params)
    print("start")
    start = time.time()
    jac_eval = mag_jac(params)
    end = time.time()
    print("jac finish: %.3f sec"%(end - start))

    t_np = np.array([t, A]).T
    jac_np = np.array(jac_eval)
    np.savetxt("example/triple-lens-jacobian/time_mag.csv", t_np, delimiter=",")
    np.save("example/triple-lens-jacobian/jacobian.npy", jac_np)

# Plotting
file = np.loadtxt("example/triple-lens-jacobian/time_mag.csv", delimiter=",")
t, A = file.T[0], file.T[1]
jac = np.load("example/triple-lens-jacobian/jacobian.npy").T

param_names = ['t0', 'tE', 'u0', 'q', 's', 'alpha', 'rho', 'q3', 'r3', 'psi']

n_params = jac.shape[0]

fig, axes = plt.subplots(11, 1, figsize=(12, 10), sharex=True,
                         gridspec_kw={'height_ratios': [8, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], 'wspace':0.3}) 
                         #gridspec_kw={'hspace': 0.1, 'height_ratios': [2] + [1]*n_params})

axes[0].plot(t, A, label='Magnification $A(t)$', color='black')
axes[0].set_ylabel('Magnification')
#axes[0].legend(loc='upper right')

from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.ticker import AutoMinorLocator
ax_in = inset_axes(axes[0],
    width="70%", # 
    height="70%", 
    bbox_transform=axes[0].transAxes,
    bbox_to_anchor=(0.05, 0.05, .9, .9),
    #bbox_to_anchor=(-0.45, 0.05, .9, .9),
)
ax_in.set_aspect(1)
ax_in.set_aspect(1)
ax_in.set(xlabel="Re$(w)$", ylabel="Im$(w)$")

s  = 1.1  # separation between the two lenses in units of total ang. Einstein radii
ax_in.set_aspect(1)
ax_in.set_aspect(1)
ax_in.set(xlabel="Re$(w)$", ylabel="Im$(w)$")

_params = {"q": q, "s": s, "q3": q3, "r3": jnp.abs(r3_complex), "psi": psi}

critical_curves, caustic_curves = critical_and_caustic_curves(npts=1000, nlenses=3, **_params)
for cc in caustic_curves:
    ax_in.plot(cc.real, cc.imag, color='red', lw=0.7)
for cc in critical_curves:
    ax_in.plot(cc.real, cc.imag, color='green', lw=0.7)
ax_in.plot(-q*s, 0 ,"x",c="k", ms=2)
ax_in.plot((1.0-q)*s, 0 ,"x",c="k", ms=2)
ax_in.plot(r3_complex.real - (0.5*s - s/(1 + q)), r3_complex.imag ,"x",c="k", ms=2)

circles = [
    plt.Circle((xi,yi), radius=rho, fill=False, facecolor=None, zorder=-1) for xi,yi in zip(w_points.real, w_points.imag)
]
c = mpl.collections.PatchCollection(circles, 
                                    match_original=True, 
                                    alpha=0.05, 
                                    edgecolor="blue", 
                                    linewidth=0.5, 
                                    zorder=10)
ax_in.add_collection(c)
ax_in.set_aspect(1)
ax_in.set(xlim=(-1.5, 1.5), ylim=(-1.5, 1.5))

labels = [
    r'$\frac{\partial A}{\partial u_0}$', r'$\frac{\partial A}{\partial t_0}$',r'$\frac{\partial A}{\partial t_E}$',
    r'$\frac{\partial A}{\partial q}$', r'$\frac{\partial A}{\partial s}$', r'$\frac{\partial A}{\partial \alpha}$',
    r'$\frac{\partial A}{\partial \rho}$',
    r'$\frac{\partial A}{\partial q_3}$', r'$\frac{\partial A}{\partial r_3}$', r'$\frac{\partial A}{\partial \psi}$'
]
for i, l in enumerate(labels):
    axes[i+1].plot(t, jac[i])
    axes[i+1].set_ylabel(l)

axes[-1].set_xlabel('Time (day)')

plt.tight_layout()
plt.savefig("example/triple-lens-jacobian/full_jacobian_plot.png", dpi=300)
plt.show()


import numpy as np

from microjax.fastlens.mag_fft import magnification_limb


def test_mag_limb_A0_matches_A_at_zero_for_large_rho():
    # n_limb=1 and 2 for coverage
    for n in (1, 2):
        rho = 1e-2
        mag = magnification_limb(n_limb=n)
        a0 = mag.A0(rho)
        a = mag.A(0.0, rho)
        assert np.allclose(a, a0, rtol=1e-10, atol=1e-12)


def test_mag_limb_small_rho_point_source_limit():
    rho = 1e-6
    u = 1e2 * rho
    mag = magnification_limb(n_limb=1)
    # For small rho and large u relative to rho, A approaches 1 (point source ~1 at large u)
    a = mag.A(u, rho)
    assert np.all(a > 1.0)
    # As u increases further, approach 1
    a2 = mag.A(3e2 * rho, rho)
    assert (a2 - 1.0) < (a - 1.0)




import jax
import jax.numpy as jnp
from microjax.fastlens.fftlog_jax import fftlog, hankel
from microjax.fastlens.special import gamma, j0, j1, j2, j1p5
from microjax.fastlens.special import ellipk, ellipe, ellipk_, ellipe_
from jax import jit, lax, vmap

def A_point(u):
    return (u**2 + 2) / jnp.abs(u) / (u**2 + 4)**0.5

class mag_:
    def __init__(self, fft_logumin=-6, fft_logumax=3, N_fft=1024, normalize_sk=True, rho_switch=1e-4, u_switch=10):
        """
        Args:
            fft_logumin      (int): minimum of FFT bin in u space (default=-6)
            fft_logumax      (int): maximum of FFT bin in u space (default=3)
            N_fft            (int): number of FFT bin in u space  (default=1024)
            normalize_sk    (bool): flag to normalize source profile in fourier space
        """
        self.fft_logumin = fft_logumin
        self.fft_logumax = fft_logumax
        self.N_fft = N_fft
        self.init_Apk()

    def init_Apk(self):
        """
        Initializing the FFT counter part of point-source magnification
        """
        u = jnp.logspace(self.fft_logumin, self.fft_logumax, self.N_fft)
        u2Au = ((u**2 + 2.0) / (u**2 + 4.0)**0.5 / u - 1) * u**2
        h = hankel(u, u2Au, nu=1.5)
        #h = hankel(u, u2Au, nu=1.5, N_extrap_high=512, N_extrap_low=512)
        self.k, apk = h.hankel(0)
        self.apk = apk * 2 * jnp.pi
    
    def sk(self, k, rho, a1):
        # Implement Fourier counter part of source profile.
        pass

    def A0(self, rho, a1):
        # Implement A(0, rho).
        pass

    def A(self, u, rho, a1=0.0):
        """
        Evaluates the extended-source magnification for rho large enough
        Aargs:
            u   (array)
            rho (float)
            a1  (float): limb darkening coefficient
        """
        u = jnp.atleast_1d(jnp.abs(u))
        a_base = jnp.ones(u.shape) * self.A0(rho, a1)
        # typical scale of source profile
        k_rho = 2 * jnp.pi / rho
        # dumping factor to avoid noisy result
        dump = jnp.exp(-(self.k / k_rho / 20)**2)
        # Fourier counter part of extended-source magnification
        cj = self.apk * self.k**2 * self.sk(self.k, rho, a1) * dump
        # Hankel back the extended-source magnification
        h = hankel(self.k, cj, nu=1.5, N_pad=512)
        u_fft, a_fft = h.hankel(0)
        a_fft = a_fft / 2.0 / jnp.pi
        a_fft = a_fft + 1.0
        # Truncate the result u>100 and append A(u=100)=1
        max_u_fft = 100
        u_fft_truncated = jnp.where(u_fft < max_u_fft, u_fft, max_u_fft)
        a_fft_truncated = jnp.where(u_fft < max_u_fft, a_fft, 1)
        log_u = jnp.log10(u)
        log_u_fft_truncated = jnp.log10(u_fft_truncated)
        interp_values = jnp.interp(log_u, log_u_fft_truncated, a_fft_truncated, right=1)
        a = jnp.where(u > 0, interp_values, a_base)

        return a 

class mag_disk(mag_):
    def sk(self, k, rho, a1):
        k = jnp.atleast_1d(k)
        x = k * rho
        idx = x > 0
        a = jnp.where(idx, 2 * j1(x) / x, jnp.ones_like(x))
        return a

    def A0(self, rho, a1):
        return (rho**2 + 4)**0.5 / rho

class mag_limb(mag_):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def su(self, u, rho, a1):
        su = jnp.zeros(u.size)
        u_rho = jnp.where(u < rho, u / rho, 1.0)
        #norm = (1.0 - a1) * jnp.pi * rho**2
        su = jnp.where(u < rho, 1.0 - a1 * (1.0 - jnp.sqrt(1.0 - u_rho**2)), su)
        #su = su / norm
        return su

    def sk(self, k, rho, a1):
        u = jnp.logspace(self.fft_logumin, self.fft_logumax, self.N_fft)
        u2su = u**2 * self.su(u, rho, a1)
        h = hankel(u, u2su, nu=1.5)
        k, sk = h.hankel(0)
        sk = sk/sk[0]
        return sk

    def A0(self, rho, a1):
        A0_disk  = (rho**2 + 4)**0.5 / rho 
        A0_term1 = (2 + 1) * (2 * (rho**2 + 2) * ellipe(-rho**2 / 4) - (rho**2 + 4) * ellipk(-rho**2 / 4)) / 3.0 / rho**3 
        norm = 1.0 - a1
        return (A0_disk - a1 * A0_term1) / norm 

# FFT based magnification
class magnification:
    def __init__(self, fft_logumin=-6, fft_logumax=3, N_fft=1024, normalize_sk=True, rho_switch=1e-4, u_switch=10):
        """
        Args:
            fft_logumin      (int): minimum of FFT bin in u space (default=-6)
            fft_logumax      (int): maximum of FFT bin in u space (default=3)
            N_fft            (int): number of FFT bin in u space  (default=1024)
            normalize_sk    (bool): flag to normalize source profile in fourier space
        """
        # Defining FFT bin
        # Default choice is validated for rho > 1e-4 to ensure 0.3% precision
        self.fft_logumin = fft_logumin
        self.fft_logumax = fft_logumax
        self.N_fft = N_fft
        self.normalize_sk = normalize_sk

        # The scale for rho and u to switch to use the approximate solution.
        # When rho < rho_switch, we use the approximate solution.
        # For u < u_switch*rho, we use Eq. (13) which is precomputed by init_small_rho below.
        # For u > u_switch*rho, we use point-source magnification.
        self.rho_switch = rho_switch
        self.u_switch = u_switch

        # initialization
        self.init_Apk()
        self.init_Aext0()
        
    def init_Apk(self):
        """
        Initializing the FFT counter part of point-source magnification
        """
        u = jnp.logspace(self.fft_logumin, self.fft_logumax, self.N_fft)
        u2Au = ((u**2 + 2.0) / (u**2 + 4.0)**0.5 / u - 1) * u**2
        h = hankel(u, u2Au, nu=1.5)
        #h = hankel(u, u2Au, nu=1.5, N_extrap_high=512, N_extrap_low=512)
        self.k, apk = h.hankel(0)
        self.apk = apk * 2 * jnp.pi

    def init_Aext0(self):
        """
        Implementation of A_ext0 in Eq. (13)
        """
        x = jnp.logspace(self.fft_logumin, self.fft_logumax, self.N_fft)
        #x = jnp.logspace(-5, 5, 1024)
        dump = jnp.exp(-(x / 100)**2)
        #min_value = 1e-12
        #dump = jnp.where(dump < min_value, min_value, dump)
        fx = x * self.sk(x, 1) * dump
        #print("init_Aext0:", x, fx, self.sk(x, 1), dump)
        h = hankel(x, fx, nu=1.5, N_pad=self.N_fft)
        #h = hankel(x, fx, nu=1.5, N_pad=1024,)
        u, aext0 = h.hankel(0)
        self.Aext0 = lambda x: jnp.interp(x, u, aext0)

    def sk(self, k, rho):
        # Implement Fourier counter part of source profile.
        pass

    def A0(self, rho):
        # Implement A(0, rho).
        pass

    def _A_for_small_rho(self, u, rho):
        """
        Evaluates the extended-source magnification for small rho.
        Aargs:
            u   (array)
            rho (float)
        """
        u = jnp.atleast_1d(jnp.abs(u))
        # Assign approximated solution: Eq. (13)
        a = jnp.ones(u.shape)
        idx = u < self.u_switch * rho
        x = jnp.where(idx, u / rho, jnp.ones_like(u))
        val = jnp.where(idx, self.Aext0(x) / rho + 1, jnp.ones_like(u))
        # Assign approximated solution: point-source magnification
        a = jnp.where(idx, val, A_point(u))
        #a = jnp.where(~idx, A_point(u), a)
        return a

    def _A_for_large_rho(self, u, rho):
        """
        Evaluates the extended-source magnification for rho large enough
        Aargs:
            u   (array)
            rho (float)
        """
        u = jnp.atleast_1d(jnp.abs(u))
        a_base = jnp.ones(u.shape) * self.A0(rho)
        # typical scale of source profile
        k_rho = 2 * jnp.pi / rho
        # dumping factor to avoid noisy result
        dump = jnp.exp(-(self.k / k_rho / 20)**2)
        # Fourier counter part of extended-source magnification
        cj = self.apk * self.k**2 * self.sk(self.k, rho) * dump
        # Hankel back the extended-source magnification
        h = hankel(self.k, cj, nu=1.5, N_pad=512)
        u_fft, a_fft = h.hankel(0)
        a_fft = a_fft / 2 / jnp.pi
        a_fft = a_fft + 1
        # Truncate the result u>100 and append A(u=100)=1
        max_u_fft = 100
        u_fft_truncated = jnp.where(u_fft < max_u_fft, u_fft, max_u_fft)
        a_fft_truncated = jnp.where(u_fft < max_u_fft, a_fft, 1)
        log_u = jnp.log10(u)
        log_u_fft_truncated = jnp.log10(u_fft_truncated)
        interp_values = jnp.interp(log_u, log_u_fft_truncated, a_fft_truncated, right=1)
        a = jnp.where(u > 0, interp_values, a_base)

        return a
    
    def A(self, u, rho):
        """
        Returns the extended-source magnification of microlensing light-curve.
        Args:
            u   (float): impact parameter in the lens plane normalized by Einstein angle.
            rho (float): source scale parameter.
        Returns:
            a   (float): magnification for extended-source profile.
        """
        u = jnp.atleast_1d(jnp.abs(u))
        """
        def small(_):
            return self._A_for_small_rho(u, rho)
        def large(_):
           return self._A_for_large_rho(u, rho) 
        return lax.cond(rho < self.rho_switch, small, large, None) 
        """
        small_rho = self._A_for_small_rho(u, rho) 
        large_rho = self._A_for_large_rho(u, rho) 
        return jnp.where(rho < self.rho_switch, small_rho, large_rho) 

class magnification_disk(magnification):
    def sk(self, k, rho):
        k = jnp.atleast_1d(k)
        x = k * rho
        idx = x > 0
        a = jnp.where(idx, 2 * j1(x) / x, jnp.ones_like(x))
        return a

    def A0(self, rho):
        return (rho**2 + 4)**0.5 / rho

class mag_limb1(magnification):
    def __init__(self, a1=0.5, **kwargs):
        self.a1 = a1 #limb-darkening coeff
        super().__init__(**kwargs)

    def su(self, u, rho, a1=None):
        su = jnp.zeros(u.size)
        su = jnp.where(u < rho, 1.0 - self.a1 * (1.0 - jnp.sqrt(jnp.abs(1.0 - u**2 / rho**2))), su)
        su = su/su[0] 
        return su

    def sk(self, k, rho):
        u = jnp.logspace(self.fft_logumin, self.fft_logumax, self.N_fft)
        u2su = u**2 * self.su(u, rho)
        h = hankel(u, u2su, nu=1.5)
        #h = hankel(u, u2su, nu=1.5)
        k, sk = h.hankel(0)
        sk = sk/sk[0]
        return sk

    def sk__(self, k, rho):
        k = jnp.atleast_1d(k)
        x = k * rho
        sk_disk  = 2.0 * j1(x) / x
        sk_term1 = 3.0 * (jnp.sin(x) - x * jnp.cos(x)) / x**3
        norm = 1.0 - self.a1
        sk = jnp.where(x > 0, (sk_disk - self.a1 * sk_term1) / norm, jnp.ones_like(x))
        return sk 


    def sk_(self, k, rho):
        k = jnp.atleast_1d(k)
        x = k * rho
        nu = 1.5
        a_base = jnp.ones(x.shape) * 1.0 / (1 + 2)
        norm = 1.0 - self.a1
        sk_disk  = 2 * j1(x) / x 
        sk_term1 = nu * 2**nu * gamma(nu) * j1p5(x) / x**nu 
        sk = jnp.where(x > 0, (sk_disk - self.a1 * sk_term1) / norm, a_base)
        #sk = sk / sk[0]
        return sk

    def A0(self, rho):
        A0_disk  = (rho**2 + 4)**0.5 / rho 
        A0_term1 = (2 + 1) * (2 * (rho**2 + 2) * ellipe(-rho**2 / 4) - (rho**2 + 4) * ellipk(-rho**2 / 4)) / 3.0 / rho**3 
        norm = 1.0 - self.a1
        return (A0_disk - self.a1 * A0_term1) / norm 


class magnification_limb1(magnification):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def sk(self, k, rho):
        k = jnp.atleast_1d(k)
        x = k * rho
        nu = 1.5
        a_base = jnp.ones(x.shape)*1.0 / (1 + 2)
        sk = jnp.where(x > 0, nu * 2**nu * gamma(nu) * j1p5(x) / x**nu, a_base)
        return sk

    def A0(self, rho): 
        return (2 + 1) * (2 * (rho**2 + 2) * ellipe(-rho**2 / 4) - (rho**2 + 4) * ellipk(-rho**2 / 4)) / 3.0 / rho**3

class magnification_limb2(magnification):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def sk(self, k, rho):
        k = jnp.atleast_1d(k)
        x = k * rho
        nu = 2
        a_base = jnp.ones(x.shape)*1.0 / (2 + 2)
        a = jnp.where(x > 0, 2**nu * gamma(nu) * j2(x) / x**nu * nu, a_base)
        return a

    def A0(self, rho):
        return (2 + 2) * (rho * (rho**2 + 2) * (rho**2 + 4)**0.5 - 8*jnp.arcsinh(rho/2)) / 4/rho**4

class magnification_log(magnification):
    def su(self, u, rho):
        x = u / rho
        ans = jnp.zeros(x.shape)
        idx = x < 1
        sq = jnp.sqrt(1 - x[idx]**2)
        ans = ans.at[idx].set(sq * jnp.log(sq))
        return x * jnp.log(x)

    def sk(self, k, rho):
        pass

    def A0(self, rho):
        pass



"""
python module for calculating integrals with 1 Bessel functions.
This module contains:
-- FFTLog method for integrals with 1 spherical Bessel function;
-- integrals with 1 (cylindrical) Bessel function, i.e. Hankel transform;
-- window function (optional) for smoothing Fourier coefficients

by Xiao Fang
Apr 10, 2019
"""

import numpy as np
from scipy.special import gamma
from numpy.fft import rfft, irfft

class fftlog(object):

	def __init__(self, x, fx, nu=1.1, N_extrap_low=0, N_extrap_high=0, c_window_width=0.25, N_pad=0):

		self.x_origin = x # x is logarithmically spaced
		# self.lnx = np.log(x)
		self.dlnx = np.log(x[1]/x[0])
		self.fx_origin= fx # f(x) array
		self.nu = nu
		self.N_extrap_low = N_extrap_low
		self.N_extrap_high = N_extrap_high
		self.c_window_width = c_window_width

		# extrapolate x and f(x) linearly in log(x), and log(f(x))
		self.x = log_extrap(x, N_extrap_low, N_extrap_high)
		self.fx = log_extrap(fx, N_extrap_low, N_extrap_high)
		self.N = self.x.size

		# zero-padding
		self.N_pad = N_pad
		if(N_pad):
			pad = np.zeros(N_pad)
			self.x = log_extrap(self.x, N_pad, N_pad)
			self.fx = np.hstack((pad, self.fx, pad))
			self.N += 2*N_pad
			self.N_extrap_high += N_pad
			self.N_extrap_low += N_pad

		if(self.N%2==1): # Make sure the array sizes are even
			self.x= self.x[:-1]
			self.fx=self.fx[:-1]
			self.N -= 1
			if(N_pad):
				self.N_extrap_high -=1

		self.m, self.c_m = self.get_c_m()
		self.eta_m = 2*np.pi/(float(self.N)*self.dlnx) * self.m


	def get_c_m(self):
		"""
		return m and c_m
		c_m: the smoothed FFT coefficients of "biased" input function f(x): f_b = f(x) / x^\nu
		number of x values should be even
		c_window_width: the fraction of c_m elements that are smoothed,
		e.g. c_window_width=0.25 means smoothing the last 1/4 of c_m elements using "c_window".
		"""

		f_b=self.fx * self.x**(-self.nu)
		c_m=rfft(f_b)
		m=np.arange(0,self.N//2+1) 
		c_m = c_m*c_window(m, int(self.c_window_width*self.N//2.) )
		return m, c_m

	def fftlog(self, ell):
		"""
		Calculate F(y) = \int_0^\infty dx / x * f(x) * j_\ell(xy),
		where j_\ell is the spherical Bessel func of order ell.
		array y is set as y[:] = (ell+1)/x[::-1]
		"""
		x0 = self.x[0]
		z_ar = self.nu + 1j*self.eta_m
		y = (ell+1.) / self.x[::-1]
		h_m = self.c_m * (self.x[0]*y[0])**(-1j*self.eta_m) * g_l(ell, z_ar)

		Fy = irfft(np.conj(h_m)) * y**(-self.nu) * np.sqrt(np.pi)/4.
		#print(self.N_extrap_high,self.N,self.N_extrap_low)
		return y[self.N_extrap_high:self.N-self.N_extrap_low], Fy[self.N_extrap_high:self.N-self.N_extrap_low]

	def fftlog_dj(self, ell):
		"""
		Calculate F(y) = \int_0^\infty dx / x * f(x) * j'_\ell(xy),
		where j_\ell is the spherical Bessel func of order ell.
		array y is set as y[:] = (ell+1)/x[::-1]
		"""
		x0 = self.x[0]
		z_ar = self.nu + 1j*self.eta_m
		y = (ell+1.) / self.x[::-1]
		h_m = self.c_m * (self.x[0]*y[0])**(-1j*self.eta_m) * g_l_1(ell, z_ar)

		Fy = irfft(np.conj(h_m)) * y**(-self.nu) * np.sqrt(np.pi)/4.
		return y[self.N_extrap_high:self.N-self.N_extrap_low], Fy[self.N_extrap_high:self.N-self.N_extrap_low]

	def fftlog_ddj(self, ell):
		"""
		Calculate F(y) = \int_0^\infty dx / x * f(x) * j''_\ell(xy),
		where j_\ell is the spherical Bessel func of order ell.
		array y is set as y[:] = (ell+1)/x[::-1]
		"""
		x0 = self.x[0]
		z_ar = self.nu + 1j*self.eta_m
		y = (ell+1.) / self.x[::-1]
		h_m = self.c_m * (self.x[0]*y[0])**(-1j*self.eta_m) * g_l_2(ell, z_ar)

		Fy = irfft(np.conj(h_m)) * y**(-self.nu) * np.sqrt(np.pi)/4.
		return y[self.N_extrap_high:self.N-self.N_extrap_low], Fy[self.N_extrap_high:self.N-self.N_extrap_low]

	def fftlog_jsqr(self, ell):
		"""
		Calculate F(y) = \int_0^\infty dx / x * f(x) * (j_\ell(xy))^2,
		where j_\ell is the spherical Bessel func of order ell.
		array y is set as y[:] = (ell+1)/x[::-1]
		"""
		x0 = self.x[0]
		z_ar = self.nu + 1j*self.eta_m
		y = (ell+1.) / self.x[::-1]
		h_m = self.c_m * (self.x[0]*y[0])**(-1j*self.eta_m) * h_l(ell, z_ar)

		Fy = irfft(np.conj(h_m)) * y**(-self.nu) * np.sqrt(np.pi)/4.
		#print(self.N_extrap_high,self.N,self.N_extrap_low)
		return y[self.N_extrap_high:self.N-self.N_extrap_low], Fy[self.N_extrap_high:self.N-self.N_extrap_low]



class hankel(object):
	def __init__(self, x, fx, nu, N_extrap_low=0, N_extrap_high=0, c_window_width=0.25, N_pad=0):
		#print('nu is required to be between (0.5-n) and 2.')
		self.myfftlog = fftlog(x, np.sqrt(x)*fx, nu, N_extrap_low, N_extrap_high, c_window_width, N_pad)
	
	def hankel(self, n):
		y, Fy = self.myfftlog.fftlog(n-0.5)
		Fy *= np.sqrt(2*y/np.pi)
		return y, Fy



### Utility functions ####################

def log_extrap(x, N_extrap_low, N_extrap_high):

	low_x = high_x = []
	if(N_extrap_low):
		dlnx_low = np.log(x[1]/x[0])
		low_x = x[0] * np.exp(dlnx_low * np.arange(-N_extrap_low, 0) )
	if(N_extrap_high):
		dlnx_high= np.log(x[-1]/x[-2])
		high_x = x[-1] * np.exp(dlnx_high * np.arange(1, N_extrap_high+1) )
	x_extrap = np.hstack((low_x, x, high_x))
	return x_extrap

def c_window(n,n_cut):
	"""
	One-side window function of c_m,
	Adapted from Eq.(C1) in McEwen et al. (2016), arXiv:1603.04826
	"""
	n_right = n[-1] - n_cut
	# n_left = n[0]+ n_cut 
	n_r=n[ n[:]  > n_right ] 
	# n_l=n[ n[:]  <  n_left ] 
	theta_right=(n[-1]-n_r)/float(n[-1]-n_right-1) 
	# theta_left=(n_l - n[0])/float(n_left-n[0]-1) 
	W=np.ones(n.size)
	W[n[:] > n_right]= theta_right - 1/(2*np.pi)*np.sin(2*np.pi*theta_right)
	# W[n[:] < n_left]= theta_left - 1/(2*pi)*sin(2*pi*theta_left)
	return W

def g_m_vals(mu,q):
	'''
	g_m_vals function is adapted from FAST-PT
	g_m_vals(mu,q) = gamma( (mu+1+q)/2 ) / gamma( (mu+1-q)/2 ) = gamma(alpha+)/gamma(alpha-)
	mu = (alpha+) + (alpha-) - 1
	q = (alpha+) - (alpha-)

	switching to asymptotic form when |Im(q)| + |mu| > cut = 200
	'''
	if(mu+1+q.real[0]==0):
		print("gamma(0) encountered. Please change another nu value! Try nu=1.1 .")
		exit()
	imag_q= np.imag(q)
	g_m=np.zeros(q.size, dtype=complex)
	cut =200
	asym_q=q[np.absolute(imag_q)+ np.absolute(mu) >cut]
	asym_plus=(mu+1+asym_q)/2.
	asym_minus=(mu+1-asym_q)/2.

	q_good=q[ (np.absolute(imag_q)+ np.absolute(mu) <=cut) & (q!=mu + 1 + 0.0j)]

	alpha_plus=(mu+1+q_good)/2.
	alpha_minus=(mu+1-q_good)/2.
	
	g_m[(np.absolute(imag_q)+ np.absolute(mu) <=cut) & (q!= mu + 1 + 0.0j)] =gamma(alpha_plus)/gamma(alpha_minus)

	# asymptotic form 								
	g_m[np.absolute(imag_q)+ np.absolute(mu)>cut] = np.exp( (asym_plus-0.5)*np.log(asym_plus) - (asym_minus-0.5)*np.log(asym_minus) - asym_q \
		+1./12 *(1./asym_plus - 1./asym_minus) +1./360.*(1./asym_minus**3 - 1./asym_plus**3) +1./1260*(1./asym_plus**5 - 1./asym_minus**5) )

	g_m[np.where(q==mu+1+0.0j)[0]] = 0.+0.0j
	return g_m

def g_m_ratio(a):
	'''
	g_m_ratio(a) = gamma(a)/gamma(a+0.5)
	switching to asymptotic form when |Im(a)| > cut = 200
	'''
	if(a.real[0]==0):
		print("gamma(0) encountered. Please change another nu value! Try nu=1.1 .")
		exit()
	imag_a= np.imag(a)
	g_m=np.zeros(a.size, dtype=complex)
	cut =100
	asym_a=a[np.absolute(imag_a) >cut]
	asym_a_plus = asym_a+0.5

	a_good=a[ (np.absolute(imag_a) <=cut)]
	
	g_m[(np.absolute(imag_a) <=cut)] =gamma(a_good)/gamma(a_good+0.5)

	# asymptotic form 								
	g_m[np.absolute(imag_a)>cut] = np.exp( (asym_a-0.5)*np.log(asym_a) - asym_a*np.log(asym_a_plus) + 0.5 \
		+1./12 *(1./asym_a - 1./asym_a_plus) +1./360.*(1./asym_a_plus**3 - 1./asym_a**3) +1./1260*(1./asym_a**5 - 1./asym_a_plus**5) )

	# g_m[np.where(q==mu+1+0.0j)[0]] = 0.+0.0j
	return g_m

def g_l(l,z_array):
	'''
	gl = 2^z_array * gamma((l+z_array)/2.) / gamma((3.+l-z_array)/2.)
	alpha+ = (l+z_array)/2.
	alpha- = (3.+l-z_array)/2.
	mu = (alpha+) + (alpha-) - 1 = l+0.5
	q = (alpha+) - (alpha-) = z_array - 1.5
	'''
	gl = 2.**z_array * g_m_vals(l+0.5,z_array-1.5)
	return gl

def g_l_1(l,z_array):
	'''
	for integral containing one first-derivative of spherical Bessel function
	gl1 = -2^(z_array-1) *(z_array -1)* gamma((l+z_array-1)/2.) / gamma((4.+l-z_array)/2.)
	mu = l+0.5
	q = z_array - 2.5
	'''
	gl1 = -2.**(z_array-1) *(z_array -1) * g_m_vals(l+0.5,z_array-2.5)
	return gl1

def g_l_2(l,z_array):
	'''
	for integral containing one 2nd-derivative of spherical Bessel function
	gl2 = 2^(z_array-2) *(z_array -1)*(z_array -2)* gamma((l+z_array-2)/2.) / gamma((5.+l-z_array)/2.)
	mu = l+0.5
	q = z_array - 3.5
	'''
	gl2 = 2.**(z_array-2) *(z_array -1)*(z_array -2)* g_m_vals(l+0.5,z_array-3.5)
	return gl2

def h_l(l,z_array):
	'''
	hl = gamma(l+ z_array/2.) * gamma((2.-z_array)/2.) / gamma((3.-z_array)/2.) / gamma(2.+l -z_array/2.)
	first component is g_m_vals(2l+1, z_array - 2)
	second component is gamma((2.-z_array)/2.) / gamma((3.-z_array)/2.)
	'''
	hl = g_m_vals(2*l+1., z_array - 2.) * g_m_ratio((2.-z_array)/2.)
	return hl


import numpy as np
import jax.numpy as jnp
import pytest

pytest.importorskip("VBMicrolensing")

from microjax.trajectory.parallax import (
    peri_vernal,
    set_parallax,
    compute_parallax,
)


def compute_microjax_pspl_parallax(t, u0, t0, tE, piEN, piEE, RA, Dec):
    tref = t0
    tperi, tvernal = peri_vernal(tref)
    parallax_params = set_parallax(tref, tperi, tvernal, RA, Dec)
    dtn, dum = compute_parallax(t, piEN, piEE, parallax_params)
    tau = (t - t0) / tE
    um = u0 + dum
    tm = tau + dtn
    u2 = um**2 + tm**2
    u = jnp.sqrt(u2)
    A = (u2 + 2.0) / (u * jnp.sqrt(u2 + 4.0))
    return A, tm, um


@pytest.mark.parametrize(
    "coords,u0,tE,t0,piEN,piEE",
    [
        ("17:45:40 -29:00:28", 0.01, 30.0, 8000.0, 0.1, 0.1),
        ("17:45:40 -29:00:28", 0.30, 20.0, 7500.0, -0.2, 0.05),
    ],
)
def test_pspl_parallax_matches_vbbl(coords, u0, tE, t0, piEN, piEE):
    import VBMicrolensing
    from astropy.coordinates import SkyCoord
    import astropy.units as u

    c = SkyCoord(coords, frame="icrs", unit=(u.hourangle, u.deg))
    RA = c.ra.deg
    Dec = c.dec.deg

    t = t0 + np.linspace(-5.0 * tE, 5.0 * tE, 200)
    log_tE = np.log(tE)

    # VBBinaryLensing via VBMicrolensing wrapper
    VBM = VBMicrolensing.VBMicrolensing()
    VBM.parallaxsystem = 1
    VBM.t0_par_fixed = 1
    VBM.t0_par = t0
    VBM.SetObjectCoordinates(coords)
    VBM.RelTol = 1e-04
    VBM.Tol = 1e-04
    param = [u0, log_tE, t0, piEN, piEE]
    mag_vbbl, x, y = VBM.PSPLLightCurveParallax(param, t)

    # microJAX implementation
    mag_jax, tm, um = compute_microjax_pspl_parallax(t, u0, t0, tE, piEN, piEE, RA, Dec)

    # Compare magnifications
    diff_mag = np.array(mag_vbbl) - np.array(mag_jax)
    assert np.max(np.abs(diff_mag)) < 5e-4

    # Compare source-plane trajectories (sign conventions differ; use -x, -y per module __main__ demo)
    x_vbbl, y_vbbl = -np.array(x), -np.array(y)
    diff_pos = np.sqrt((x_vbbl - np.array(tm)) ** 2 + (y_vbbl - np.array(um)) ** 2)
    # Position paths can differ slightly in convention/numerics; allow small tolerance
    assert np.max(diff_pos) < 2e-3



# -*- coding: utf-8 -*-
__all__ = []

__version__ = "0.1.0"
__uri__ = ''
__author__ = 'Shota Miyazaki and collaborators'
__email__ = 'miyazaki@ir.isas.jaxa.jp'
__license__ = 'MIT'
__description__ = 'Auto-differentiable Microlensing Modeling Framework with JAX.'


"""Region construction for inverse-ray polar integration in image space.

This module identifies and refines radial and angular subregions where images
of the source limb appear, to focus sampling near caustics and reduce wasted
integration. It clusters the mapped limb points and expands ranges by margins.
"""

import jax 
import jax.numpy as jnp
from jax import jit, lax, vmap
from functools import partial
from microjax.point_source import lens_eq, _images_point_source
from typing import Tuple

Array = jnp.ndarray

def define_regions(
    image_limb: Array,
    mask_limb: Array,
    rho: float,
    bins_r: int = 50,
    bins_th: int = 120,
    margin_r: float = 0.5,
    margin_th: float = 0.5,
    nlenses: int = 2,
) -> Tuple[Array, Array]:
    """Determine polar subregions for inverse-ray integration.

    Clusters the radii and angles of image points mapped from the source limb
    and builds minimal rectangles in ``(r, theta)`` covering them, then expands
    by margins and merges wrap-around angular intervals when appropriate.
    Returns arrays of shape ``(M, 2)`` for ``r_scan`` and ``th_scan`` with the
    min/max of each region.
    """
    if nlenses == 2:
        nimages_init = 10
        nimage_real = 5
    elif nlenses == 3:
        nimages_init = 18
        nimage_real = 9
    else:
        raise ValueError("Only 2 or 3 lenses are supported.")

    image_limb = image_limb.ravel()
    mask_limb = mask_limb.ravel()

    r_limb     = jnp.abs(image_limb * mask_limb)
    th_limb    = jnp.mod(jnp.arctan2(image_limb.imag, image_limb.real), 2*jnp.pi) * mask_limb
    r_mins, r_maxs   = cluster_1d(r_limb, bins=bins_r, max_cluster=nimages_init)
    th_mins, th_maxs = cluster_1d(th_limb, bins=bins_th, mode_r=False, max_cluster=nimages_init)

    r_map  = jnp.array([r_mins, r_maxs]).T
    th_map = jnp.array([th_mins, th_maxs]).T
    th_map = jnp.where(jnp.isclose(th_map, 0.0), 0.0, th_map)
    th_map = jnp.where(jnp.isclose(th_map, 2*jnp.pi), 2*jnp.pi, th_map)

    r_map_refine = _refine_addmargin_merge_r(r_limb, r_map, margin=margin_r*rho)

    in_mask   = _compute_in_mask(r_limb, th_limb, r_map_refine, th_map)
    r_masked  = jnp.repeat(r_map_refine, th_map.shape[0], axis=0) * in_mask.ravel()[:, None]
    th_masked = jnp.tile(th_map, (r_map_refine.shape[0], 1)) * in_mask.ravel()[:, None]

    r_excess   = r_masked[jnp.argsort(jnp.isclose(r_masked[:, 1], 0.0))][:nimages_init]
    th_excess  = th_masked[jnp.argsort(jnp.isclose(th_masked[:, 1], 0.0))][:nimages_init]
    r_scan, th_scan = _merge_final(r_excess, th_excess)
    r_scan, th_scan = r_scan[:nimage_real], th_scan[:nimage_real]
    # refine intervals
    r_scan_refine, th_scan_refine = _refine_final(
        r_limb, th_limb, r_scan, th_scan, margin_r=margin_r*rho, margin_th=jnp.deg2rad(margin_th)
    )
    #return r_scan, th_scan
    return r_scan_refine, th_scan_refine

def _refine_final(
    r_limb: Array,
    th_limb: Array,
    r_scan: Array,
    th_scan: Array,
    margin_r: float = 0.0,
    margin_th: float = 0.0,
) -> Tuple[Array, Array]:
    """Refine region bounds based on actual limb samples within each region."""
    r_min, r_max = r_scan[:, 0], r_scan[:, 1] # (M,) shape
    th_min, th_max = th_scan[:, 0], th_scan[:, 1] # (M,) shape
    cond_r = (r_min[:, None] < r_limb) & (r_limb < r_max[:, None])  # (M, N) shape
    cond_flip = (-jnp.pi < th_min) & (th_min < 0.0) & (0.0 < th_max) & (th_max < jnp.pi) # (M,) shape
    cond_flip_ex = cond_flip[:, None]
    th_limb_new = jnp.where(cond_flip_ex, 
                            jnp.where(th_limb[None, :] > jnp.pi, th_limb[None, :] - 2*jnp.pi, th_limb[None, :]), 
                            th_limb[None, :]) # (M, N) shape
    
    cond_th = (th_min[:, None] < th_limb_new) & (th_limb_new < th_max[:, None]) & (th_limb_new != 0) & (r_limb[None, :] != 0)  # (M, N)
    in_mask = cond_r & cond_th  # (M, N)
    # r refine
    r_min_ref  = jnp.min(jnp.where(in_mask, r_limb[None, :], 1e+5), axis=1) # (M,)
    r_max_ref  = jnp.max(jnp.where(in_mask, r_limb[None, :], 0.0), axis=1)
    r_min_ref  = jnp.maximum(0.0, r_min_ref - margin_r)
    r_max_ref  = jnp.where(r_max_ref != 0, r_max_ref + margin_r, 0.0)
    # theta refine
    th_min_ref = jnp.min(jnp.where(in_mask, th_limb_new, 1e+5), axis=1)
    th_max_ref = jnp.max(jnp.where(in_mask, th_limb_new, -1e+5), axis=1)
    # apply or not 
    case_org = (th_min > 0.0)|(cond_flip)
    r_min_new  = jnp.where(case_org, r_min_ref, r_min)
    r_max_new  = jnp.where(case_org, r_max_ref, r_max) 
    th_min_new = jnp.where(case_org, th_min_ref, th_min) - margin_th
    th_max_new = jnp.where(case_org, th_max_ref, th_max) + margin_th
    # th_range==0 if r_range==0
    th_min_new = jnp.where((r_min_new==0)&(r_max_new==0), 0.0, th_min_new)
    th_max_new = jnp.where((r_min_new==0)&(r_max_new==0), 0.0, th_max_new)
    r_scan_refine = jnp.stack([r_min_new, r_max_new], axis=1)
    th_scan_refine = jnp.stack([th_min_new, th_max_new], axis=1)

    return r_scan_refine, th_scan_refine

def refine_final(
    r_limb: Array,
    th_limb: Array,
    r_scan: Array,
    th_scan: Array,
    margin_r: float = 0.0,
    margin_th: float = 0.0,
) -> Tuple[Array, Array]:
    M = r_scan.shape[0] # number of regions
    N = r_limb.shape[0] # number of image limb points
    arr=[]
    for r_inteval, th_interval in zip(r_scan, th_scan):
        r_min, r_max = r_inteval
        th_min, th_max = th_interval 
        case_0 = th_min > 0.0 # (N,) shape
        cond_r = (r_min < r_limb)&(r_limb < r_max) # (N,) shape
        cond_th = (th_min < th_limb)&(th_limb < th_max)&(th_limb!=0)
        in_mask = (cond_r)&(cond_th)
        th_min_   = jnp.min(jnp.where(in_mask, th_limb, 1e+10))  
        th_max_   = jnp.max(jnp.where(in_mask, th_limb, -1e+10))
        r_min_new = jnp.min(jnp.where(in_mask, r_limb, 1e+10))
        r_max_new = jnp.max(jnp.where(in_mask, r_limb, -1e+10))
        r_min_new  = jnp.maximum(0.0, r_min_new - margin_r)
        r_max_new  = r_max_new + margin_r 
        th_min_new = jnp.where(case_0, th_min_, th_min)
        th_max_new = jnp.where(case_0, th_max_, th_max)
        th_min_new = th_min_new - margin_th
        th_max_new = th_max_new + margin_th
        #arr.append([r_min, r_max, th_min_new, th_max_new])
        arr.append([r_min_new, r_max_new, th_min_new, th_max_new])
    arr = jnp.vstack(arr)
    r_scan_refine = arr[:, :2]
    th_scan_refine = arr[:, 2:]
    return r_scan_refine, th_scan_refine

def cluster_1d(arr: Array, bins: int = 100, max_cluster: int = 5, mode_r: bool = True) -> Tuple[Array, Array]:
    """Cluster 1D values via occupancy bins and return start/end edges.

    - For radii (``mode_r=True``), the bin range is fit to nonzero samples.
    - For angles, uses [0, 2pi] with light regularization of intervals.
    Returns two arrays of the same length with left/right edges of up to
    ``max_cluster`` occupied intervals.
    """
    if mode_r:
        bin_min = jnp.min(jnp.where(arr == 0, jnp.inf, arr))
        bin_max = jnp.max(jnp.where(arr == 0, -jnp.inf, arr))
    else:
        # Here I do not want to optimize the intervals so much.
        bin_min = 0
        bin_max = 2 * jnp.pi
    delta = (bin_max - bin_min) / bins
    bin_edges = jnp.linspace(bin_min - delta, bin_max + delta, bins + 3, endpoint=True) # [0]-[bin+1]
    bin_indices = jnp.digitize(arr, bin_edges) - 1  # [0]-[bin+1]
    bin_indices = jnp.clip(bin_indices, 1, bins)    # [1]-[bin]
    counts = jnp.bincount(bin_indices, length=bins + 2)
    bin_mask = counts > 0 # bins + 2 
    diff_mask = jnp.diff(bin_mask.astype(int))  #length: bin + 1
    start_mask = diff_mask == 1 
    end_mask   = diff_mask == -1

    start_edges = jnp.sort(bin_edges[1:-1] * start_mask.astype(float), descending=False)[-max_cluster:]
    end_edges   = jnp.sort(bin_edges[1:-1] * end_mask.astype(float), descending=False)[-max_cluster:] 
    return start_edges, end_edges

def _compute_in_mask(r_limb: Array, th_limb: Array, r_use: Array, th_use: Array) -> Array:
    """Compute mask of shape (M, K) for regions (r_use[m], th_use[k]) covering any limb points."""
    M = r_use.shape[0]  
    K = th_use.shape[0]  
    N = r_limb.shape[0]
    r_limb_expanded = r_limb.reshape(1, 1, N)
    th_limb_expanded = th_limb.reshape(1, 1, N)
    r_use_min = r_use[:, 0].reshape(M, 1, 1)
    r_use_max = r_use[:, 1].reshape(M, 1, 1)
    th_use_min = th_use[:, 0].reshape(1, K, 1)
    th_use_max = th_use[:, 1].reshape(1, K, 1)

    r_condition = (r_limb_expanded > r_use_min) & (r_limb_expanded < r_use_max)  # shape: (M, 1, N)
    th_condition = (th_limb_expanded > th_use_min) & (th_limb_expanded < th_use_max)  # shape: (1, K, N)
    combined_condition = r_condition & th_condition  # shape: (M, K, N)

    # condition for all the combination
    in_mask = jnp.any(combined_condition, axis=2)  # shape: (M, K)
    return in_mask

def _merge_final(r_map: Array, th_map: Array) -> Tuple[Array, Array]:
    """Merge angular intervals at 0/2pi boundaries for identical radial ranges."""
    r_mins, r_maxs   = r_map[:, 0],  r_map[:, 1]  # (N,)
    th_mins, th_maxs = th_map[:, 0], th_map[:, 1]  # (N,)
    matrix = jnp.stack([r_mins, r_maxs, th_mins, th_maxs], axis=1)  # (N, 4)

    r_min_col = matrix[:, 0:1]   # (N, 1)
    r_max_col = matrix[:, 1:2]   # (N, 1)
    th_min_col = matrix[:, 2:3]  # (N, 1)
    th_max_col = matrix[:, 3:4]  # (N, 1)
    
    same_r = (r_min_col == r_mins) & (r_max_col == r_maxs) & (~jnp.all(matrix[:, None, :] == matrix, axis=2))  # (N, N)
    cond_merge = (th_min_col == 0) & (same_r) & (th_maxs == 2 * jnp.pi) # (N, N)
    th_min_adjusted = jnp.where(cond_merge, th_mins - 2 * jnp.pi, 0.0)  # (N, N)
    th_min_new = jnp.sum(th_min_adjusted, axis=1)  # (N,)
    th_min_new = jnp.where(th_min_new == 0, th_mins, th_min_new) # (N,)

    new_matrix = jnp.stack([r_mins, r_maxs, th_min_new, th_maxs], axis=1)  # (N, 4)
    cond_vanish = jnp.any(jnp.any(same_r, axis=1)&(th_mins == 0)) & (th_maxs == 2*jnp.pi)  # (N, ) shape
    new_matrix = jnp.where(cond_vanish[:, None], 0.0, new_matrix)  # (N, 4) shape
    
    sort_order = jnp.argsort(new_matrix[:, 1] == 0)
    new_matrix = new_matrix[sort_order]
    new_r_map = new_matrix[:, :2]
    new_th_map = new_matrix[:, 2:]
    return new_r_map, new_th_map

def _refine_addmargin_merge_r(r_limb: Array, r_ranges: Array, margin: float = 0.0) -> Array:
    """Refine radial ranges to actual limb extents, add margin, and merge overlaps."""
    # refine range
    M = r_ranges.shape[0]  
    N = r_limb.shape[0]
    r_limb_expanded = r_limb.reshape(1, N)
    r_mins  = r_ranges[:, 0].reshape(M, 1)
    r_maxs  = r_ranges[:, 1].reshape(M, 1)
    cond_in = (r_limb_expanded > r_mins) & (r_limb_expanded < r_maxs)  # shape: (M, N)
    valid_mins = jnp.where(cond_in, r_limb_expanded, jnp.inf)
    valid_maxs = jnp.where(cond_in, r_limb_expanded, -jnp.inf)
    r_mins = jnp.where(jnp.any(cond_in, axis=1), jnp.min(valid_mins, axis=1), 0.0)
    r_maxs = jnp.where(jnp.any(cond_in, axis=1), jnp.max(valid_maxs, axis=1), 0.0)
    # add margin
    r_mins = jnp.maximum(0.0, r_mins - margin)
    r_maxs = jnp.where(r_maxs!=0, r_maxs + margin, 0.0)
    # merge
    rows = jnp.array([r_mins, r_maxs]).T # (N, 2) shape
    N = rows.shape[0]
    rows_new = rows.copy()
    for i in range(N-1):
        row_0 = rows_new[i]
        row_1 = rows_new[i+1]
        update = row_0[1] > row_1[0]
        def true_fn(rows):
            rows = rows.at[i].set(jnp.array([0.0, 0.0]))
            rows = rows.at[i + 1].set(jnp.array([row_0[0], row_1[1]]))
            return rows
        def false_fn(rows):
            return rows
        rows_new = lax.cond(update, true_fn, false_fn, rows_new)
    return rows_new

def calc_source_limb(
    w_center: complex,
    rho: float,
    Nlimb: int = 100,
    nlenses: int = 2,
    **_params,
) -> Tuple[Array, Array]:
    """Map uniformly sampled source-limb points to image plane and mask real images.

    Returns
    - image_limb: (nimg, Nlimb) complex array of images shifted back to the lens frame.
    - mask: boolean array indicating which images are real at each limb angle.
    """
    if nlenses == 2:
        s, q = _params["s"], _params["q"]
        a = 0.5 * s
        e1 = q / (1.0 + q)
        w_limb = w_center + jnp.array(rho * jnp.exp(1.0j * jnp.linspace(0.0, 2*jnp.pi, Nlimb)), dtype=complex)
        x_cm = a * (1.0 - q) / (1.0 + q)
        w_limb_shift = w_limb - x_cm
        _params = {"q": q, "s": s, "a": a, "e1": e1}
    elif nlenses == 3:
        s, q, q3, r3, psi = _params["s"], _params["q"], _params["q3"], _params["r3"], _params["psi"]
        a = 0.5 * s
        total_mass = 1.0 + q + q3
        e1 = q / total_mass
        e2 = 1.0 / total_mass 
        #r3 = r3 * jnp.exp(1j * psi)
        _params = {"a": a, "r3": r3, "e1": e1, "e2": e2, "q": q, "s": s, "q3": q3, "psi": psi}
    else:
        raise ValueError("Only 2 or 3 lenses are supported.")
    
    w_limb = w_center + jnp.array(rho * jnp.exp(1.0j * jnp.linspace(0.0, 2*jnp.pi, Nlimb)), dtype=complex)
    x_cm = a * (1.0 - q) / (1.0 + q)
    w_limb_shift = w_limb - x_cm
    image, mask = _images_point_source(w_limb_shift, nlenses=nlenses, **_params)
    image_limb = image + x_cm
    return image_limb, mask



"""Microlensing annual parallax utilities built on JAX.

This module provides helpers to compute Earth's projected position and the
resulting annual parallax corrections for microlensing trajectories in an
equatorial (ICRS) frame. Computations are JAX-friendly (autodiff/JIT) and use
lightweight fixed-iteration numerical methods where needed.

Functions
---------
- ``peri_vernal(tref)``: Select perihelion and vernal-equinox epochs nearest
  the reference time.
- ``getpsi(phi, ecc)``: Solve Kepler's equation for the eccentric anomaly via
  a fixed-iteration Newton method.
- ``prepare_projection_basis(rotaxis_deg, psi_offset, RA, Dec)``: Build the
  orbital→equatorial rotation and sky-plane (north/east) basis vectors.
- ``project_earth_position(t, tperi, period, ecc, R, north, east)``: Project
  Earth's Sun-centered orbit position onto the target tangent plane.
- ``set_parallax(tref, tperi, tvernal, RA, Dec, ...)``: Precompute quantities
  (position, local velocity, bases) around a reference epoch.
- ``compute_parallax(t, piEN, piEE, parallax_params)``: Compute the parallax
  offsets to add to dimensionless time (``tau``) and impact parameter (``u``).

Conventions
-----------
- Times are in JD-2450000 unless stated otherwise.
- Right ascension and declination are in degrees (ICRS).
- Tangent-plane basis is orthonormal and orthogonal to the line-of-sight (LOS).
- Default Earth orbital constants: obliquity 23.44 deg, eccentricity 0.0167,
  sidereal year 365.25636 days.

Sign Conventions
----------------
- Basis construction is right-handed: ``east = z_eq × los`` and
  ``north = los × east`` where ``z_eq`` is the equatorial north pole.
- Positive ``east`` increases right ascension; positive ``north`` increases
  declination on the sky.
- Parallax offsets apply as ``tm = tau + dtn`` and ``um = u0 + dum`` where
  ``tau = (t - t0)/tE`` and ``u0`` is the impact parameter.

Example
-------
>>> tperi, tvernal = peri_vernal(8000.0)
>>> params = set_parallax(8000.0, tperi, tvernal, RA=266.4168, Dec=-29.0078)
>>> dtn, dum = compute_parallax(8000.0 + jnp.linspace(-50, 50, 100), 0.1, 0.1, params)

References
----------
- Meeus, J., Astronomical Algorithms, 2nd ed., Willmann–Bell.
- Seidelmann, P. K. (ed.), Explanatory Supplement to the Astronomical Almanac.
- Gould, A. (2004), Resolution of the Microlens Parallax Degeneracy, ApJ, 606, 319.
"""

from typing import Tuple, Union

import jax.numpy as jnp

# Lightweight array alias for readability (consistent with inverse_ray style)
Array = jnp.ndarray

def peri_vernal(tref: Union[float, Array]) -> Tuple[Array, Array]:
    """Return perihelion and vernal-equinox times closest to ``tref``.

    This utility selects, from pre-tabulated epochs, the perihelion time and
    the vernal equinox time that are closest to the provided reference time.

    The function accepts both absolute Julian Date (JD) and JD-2450000. If the
    input is larger than ``2_450_000``, it is internally shifted by subtracting
    ``2_450_000`` so it can be compared against the tables below, which are in
    JD-2450000.

    Parameters
    ----------
    tref : float or array-like
        Reference time(s) in JD or JD-2450000.

    Returns
    -------
    tperi : float
        Perihelion time (JD-2450000), closest to ``tref``.
    tvernal : float
        Vernal equinox time (JD-2450000), closest to ``tref``.

    Notes
    -----
    - The returned values are selected by nearest-neighbor search in the
      provided tables and are not interpolated.
    - If ``tref`` is an array, the nearest entry is found based on the array
      broadcasting rules of JAX, and a single pair is returned as JAX scalars.

    Examples
    --------
    >>> tperi, tvernal = peri_vernal(2458000.0)
    >>> float(tperi) > 0 and float(tvernal) > 0
    True
    """
    tref = jnp.where(tref > 2_450_000.0, tref - 2_450_000.0, tref)
    peris = jnp.array([
        1546.70833, 1913.87500, 2277.08333, 2643.70833, 3009.25000, 3372.54167,
        3740.16667, 4104.33333, 4468.50000, 4836.12500, 5199.50000, 5565.29167,
        5931.54167, 6294.70833, 6662.00000, 7026.79167, 7390.45833, 7758.08333,
        8121.75000, 8486.70833, 8853.83333
    ])
    vernals = jnp.array([
        1623.81597, 1989.06319, 2354.30278, 2719.54167, 3084.78403, 3450.02292,
        3815.26806, 4180.50486, 4545.74167, 4910.98889, 5276.23056, 5641.47292,
        6006.71806, 6371.95972, 6737.20625, 7102.44792, 7467.68750, 7832.93611,
        8198.17708, 8563.41528, 8928.65903
    ])
    dperi = jnp.abs(peris - tref)
    imin = jnp.argmin(dperi)
    return peris[imin], vernals[imin]

def getpsi(phi: Union[float, Array], ecc: float) -> Array:
    """Solve Kepler's equation for the eccentric anomaly ``psi``.

    The equation solved is ``psi - e * sin(psi) = phi`` using a fixed small
    number of Newton–Raphson iterations (5) with an empirical initial guess.
    This routine is differentiable under JAX and works with scalars or arrays.

    Parameters
    ----------
    phi : float or jax.Array
        Mean anomaly in radians. May be scalar or array-like.
    ecc : float
        Orbital eccentricity, ``0 <= ecc < 1``.

    Returns
    -------
    psi : jax.Array
        Eccentric anomaly in radians, with the same broadcasted shape as
        ``phi``.

    Notes
    -----
    - The initial guess is ``phi + sign(sin(phi)) * 0.85 * ecc`` which works
      well for moderate eccentricities without branching.
    - The iteration count is fixed to keep control-flow JIT friendly.
    """
    psi = phi + jnp.sign(jnp.sin(phi)) * 0.85 * ecc # empirical init
    for _ in range(5):
        f = psi - ecc * jnp.sin(psi) - phi
        f_prime = 1.0 - ecc * jnp.cos(psi)
        psi -= f / f_prime
    return psi

def prepare_projection_basis(rotaxis_deg: float, psi_offset: float, RA: float, Dec: float) -> Tuple[Array, Array, Array]:
    """Build rotation and on-sky projection bases.

    Constructs the rotation matrix that maps the Sun–Earth orbital plane
    coordinates to the equatorial frame and returns the orthonormal basis
    vectors on the sky plane at the target direction: ``north`` and ``east``.

    Parameters
    ----------
    rotaxis_deg : float
        Obliquity of the ecliptic (tilt between equatorial and ecliptic
        planes) in degrees.
    psi_offset : float
        Eccentric-anomaly angle between perihelion and the vernal equinox in
        radians. This aligns the orbital x-axis with the vernal direction.
    RA : float
        Right ascension of the target in degrees (ICRS).
    Dec : float
        Declination of the target in degrees (ICRS).

    Returns
    -------
    R : jax.Array, shape (3, 3)
        Rotation matrix from orbital coordinates to equatorial coordinates.
    north : jax.Array, shape (3,)
        Unit vector pointing to celestial north on the tangent plane at the
        target position.
    east : jax.Array, shape (3,)
        Unit vector pointing to celestial east on the tangent plane.

    Notes
    -----
    The line-of-sight unit vector is derived from (RA, Dec). The returned
    ``east`` and ``north`` are orthonormal and orthogonal to the line of sight.
    """
    # orbital frame -> ecliptic frame
    # psi_offset is an angle from perihelion to vernal equinox
    # (rotate about the z-axis to align with the x-axis with the vernal)
    Rz = jnp.array([
        [jnp.cos(-psi_offset), -jnp.sin(-psi_offset), 0],
        [jnp.sin(-psi_offset),  jnp.cos(-psi_offset), 0],
        [0,                   0,                    1]
    ])
    # ecliptic -> equatorial (rotate about x-axis)
    # rotaxis is an inclunation angle from equational to ecliptic frame
    rotaxis = jnp.deg2rad(rotaxis_deg)
    Rx = jnp.array([
        [1, 0,               0],
        [0, jnp.cos(rotaxis), -jnp.sin(rotaxis)],
        [0, jnp.sin(rotaxis),  jnp.cos(rotaxis)]
    ])
    R = Rx @ Rz

    alpha, delta = jnp.deg2rad(RA), jnp.deg2rad(Dec)
    los = jnp.array([
        jnp.cos(alpha) * jnp.cos(delta),
        jnp.sin(alpha) * jnp.cos(delta),
        jnp.sin(delta)
    ])
    z_eq = jnp.array([0.0, 0.0, 1.0])
    east = jnp.cross(z_eq, los)
    east /= jnp.linalg.norm(east)
    north = jnp.cross(los, east)
    north /= jnp.linalg.norm(north)

    return R, north, east

def project_earth_position(
    t: Union[float, Array],
    tperi: float,
    period: float,
    ecc: float,
    R: Array,
    north: Array,
    east: Array,
) -> Array:
    """Project Earth's position onto the target's tangent plane.

    Computes the Sun-centered position of Earth in its (elliptical) orbit at
    time ``t`` using the eccentric anomaly and projects it onto the sky-plane
    basis defined by ``north`` and ``east``.

    Parameters
    ----------
    t : float or jax.Array
        Observation time(s) in JD-2450000; scalar or 1D array.
    tperi : float
        Time of perihelion in JD-2450000.
    period : float
        Orbital period in days (sidereal year).
    ecc : float
        Orbital eccentricity, ``0 <= ecc < 1``.
    R : jax.Array, shape (3, 3)
        Rotation matrix from orbital to equatorial frame.
    north : jax.Array, shape (3,)
        North unit vector on the tangent plane.
    east : jax.Array, shape (3,)
        East unit vector on the tangent plane.

    Returns
    -------
    q : jax.Array, shape (2, N)
        Stacked projected coordinates ``[q_north, q_east]`` where ``N`` is the
        number of time samples (``N = 1`` for scalar ``t``).

    Notes
    -----
    The orbital coordinates are computed in the orbital frame with x-axis
    toward perihelion, then rotated to the equatorial frame and projected onto
    the tangent-plane basis.
    """
    t = jnp.atleast_1d(t)
    N = t.shape[0]
    phi = 2.0 * jnp.pi * (t - tperi) / period
    psi = getpsi(phi, ecc)

    # Sun-centered position, x-axis aligning with perihelion, z-axis aligning with ecliptic north 
    x_orb = jnp.cos(psi) - ecc
    y_orb = jnp.sin(psi) * jnp.sqrt(1.0 - ecc**2)
    r_orb = jnp.array([x_orb, y_orb, jnp.zeros(N)])
    r_eq = R @ r_orb # (3, N) shape

    q_north = jnp.dot(north, r_eq)
    q_east = jnp.dot(east, r_eq)
    return jnp.array([q_north, q_east])

def set_parallax(
    tref: float,
    tperi: float,
    tvernal: float,
    RA: float,
    Dec: float,
    rotaxis_deg: float = 23.44,
    ecc: float = 0.0167,
    period: float = 365.25636,
    dt: float = 0.1,
) -> Tuple[Array, Array, Array, Array, Array, float, float, float, float]:
    """Precompute parallax parameters at a reference epoch.

    Precomputes quantities needed to evaluate the microlensing annual parallax
    signal around ``tref``. This includes the on-sky Earth position at ``tref``,
    a local linear velocity approximation (finite-difference over ``dt``), and
    the projection/rotation bases.

    If either ``tperi`` or ``tvernal`` is passed as 0, both values are
    automatically inferred using :func:`peri_vernal` at ``tref``.

    Parameters
    ----------
    tref : float
        Reference time in JD-2450000 at which the linearization is anchored.
    tperi : float
        Perihelion time in JD-2450000, or 0 to auto-select.
    tvernal : float
        Vernal equinox time in JD-2450000, or 0 to auto-select.
    RA : float
        Target right ascension in degrees (ICRS).
    Dec : float
        Target declination in degrees (ICRS).
    rotaxis_deg : float, optional
        Obliquity of the ecliptic in degrees. Default is 23.44.
    ecc : float, optional
        Orbital eccentricity of Earth. Default is 0.0167.
    period : float, optional
        Orbital period (sidereal year) in days. Default is 365.25636.
    dt : float, optional
        Time step (days) used to compute the finite-difference velocity.

    Returns
    -------
    parallax_params : tuple
        Tuple ``(qne0, vne0, R, north, east, tref, tperi, period, ecc)`` where
        each element is:
        - ``qne0``: jax.Array, shape (2,), Earth position [north, east] at ``tref``.
        - ``vne0``: jax.Array, shape (2,), approximate velocity d[q_north, q_east]/dt at ``tref``.
        - ``R``: jax.Array, shape (3, 3), rotation matrix orbital→equatorial.
        - ``north``: jax.Array, shape (3,), north basis vector.
        - ``east``: jax.Array, shape (3,), east basis vector.
        - ``tref``: float, the reference epoch.
        - ``tperi``: float, perihelion epoch used.
        - ``period``: float, orbital period used.
        - ``ecc``: float, eccentricity used.

    Notes
    -----
    The velocity is computed with a symmetric finite difference of width
    ``2*dt`` to reduce truncation error and preserve JAX differentiability.
    """
    info_0 = peri_vernal(tref)
    info = jnp.where(tperi * tvernal == 0,
                     jnp.array(info_0),
                     jnp.array([tperi, tvernal]))
    tperi, tvernal = info
    phi_offset = 2 * jnp.pi * (tvernal - tperi) / period
    psi_offset = getpsi(phi_offset, ecc)
    costh = (jnp.cos(psi_offset) - ecc) / (1 - ecc * jnp.cos(psi_offset))
    sinth = jnp.sqrt(1.0 - ecc**2) * jnp.sin(psi_offset) / (1 - ecc * jnp.cos(psi_offset))
    f_rot = jnp.mod(jnp.arctan2(sinth, costh), 2 * jnp.pi)
    R, north, east = prepare_projection_basis(rotaxis_deg, f_rot, RA, Dec)
    qne0 = project_earth_position(tref, tperi, period, ecc, R, north, east)
    qne1 = project_earth_position(tref - dt, tperi, period, ecc, R, north, east)
    qne2 = project_earth_position(tref + dt, tperi, period, ecc, R, north, east)
    vne0 = 0.5 * (qne2 - qne1) / dt
    parallax_params = (qne0, vne0, R, north, east, tref, tperi, period, ecc)
    return parallax_params

def compute_parallax(
    t: Union[float, Array],
    piEN: float,
    piEE: float,
    parallax_params: Tuple[Array, Array, Array, Array, Array, float, float, float, float],
) -> Tuple[Array, Array]:
    """Compute annual parallax offsets at times ``t``.

    Produces the parallax-induced shifts to microlensing trajectory parameters:
    ``dtn`` should be added to the dimensionless time coordinate ``tau``, and
    ``dum`` should be added to the impact parameter ``u`` (north–east frame).

    Parameters
    ----------
    t : float or jax.Array
        Time(s) in JD-2450000 at which to evaluate the parallax signal.
    piEN : float
        Parallax amplitude projected in the north direction.
    piEE : float
        Parallax amplitude projected in the east direction.
    parallax_params : tuple
        Output of :func:`set_parallax`.

    Returns
    -------
    dtn : jax.Array
        Offset to add to the dimensionless time coordinate(s) ``tau``; shape
        ``(N,)`` matching the number of time samples.
    dum : jax.Array
        Offset to add to the impact parameter coordinate(s) ``u``; shape
        ``(N,)``.

    Notes
    -----
    The mean linear motion around ``tref`` is removed using the precomputed
    velocity in ``parallax_params`` to isolate the purely annual parallax
    contribution.
    """
    qne0, vne0, R, north, east, tref, tperi, period, ecc = parallax_params
    qne = project_earth_position(t, tperi, period, ecc, R, north, east)
    dt_ref = t - tref
    # Vectorized form equivalent to the original comprehension over i in {0,1}
    qne_delta = qne - (qne0 + vne0 * dt_ref)
    dtn = piEN * qne_delta[0] + piEE * qne_delta[1]
    dum = piEN * qne_delta[1] - piEE * qne_delta[0]
    return dtn, dum


if __name__ == "__main__":
    import numpy as np
    import jax.numpy as jnp
    import jax
    jax.config.update('jax_enable_x64', True)
    import matplotlib.pyplot as plt
    import VBMicrolensing

    from astropy.coordinates import SkyCoord
    import astropy.units as u
    coords = "17:45:40 -29:00:28"
    c = SkyCoord(coords, frame="icrs", unit=(u.hourangle, u.deg),)
    RA = c.ra.deg
    Dec = c.dec.deg

    u0 = 0.01
    log_tE = np.log(30.0)
    t0 = 8000.0 
    piEN = 0.3
    piEE = 0.3
    tE = np.exp(log_tE)
    t = t0 + np.linspace(-5.0*tE, 5.0*tE, 1000)

    # VBM
    VBM = VBMicrolensing.VBMicrolensing()
    VBM.parallaxsystem = 1
    VBM.t0_par_fixed = 1
    VBM.t0_par = t0
    VBM.SetObjectCoordinates(coords)
    VBM.RelTol = 1e-04
    VBM.Tol=1e-04
    param = [u0, log_tE, t0, piEN, piEE]
    mag_vbbl, x, y = VBM.PSPLLightCurveParallax(param, t)

    # MulensModel
    import MulensModel as mm
    params = dict()
    params['t_0'] = t0 + 2450000
    params['t_0_par'] = t0 + 2450000
    params['u_0'] = u0
    params['t_E'] = tE
    params['pi_E_N'] = piEN
    params['pi_E_E'] = piEE
    my_model = mm.Model(params, coords=coords)
    my_model.parallax(earth_orbital=True, satellite=False)
    mag_mulens = my_model.get_magnification(time=t+2450000,)

    # microJAX
    tref = t0
    tperi, tvernal = peri_vernal(tref)
    parallax_params = set_parallax(tref, tperi, tvernal, RA, Dec)
    dtn, dum = compute_parallax(t, piEN, piEE, parallax_params)
    tau = (t - t0)/tE
    um  = u0 + dum
    tm  = tau + dtn
    u2  = um**2 + tm**2
    u   = jnp.sqrt(u2)
    mag_jax = (u2 + 2.0) / (u * jnp.sqrt(u2 + 4.0))

    diff_vbbl_jax    = jnp.array(mag_vbbl) - jnp.array(mag_jax)
    diff_mulens_jax  = jnp.array(mag_mulens) - jnp.array(mag_jax)

    print("mag diff max VBM - MicroJAX:    ", jnp.max(diff_vbbl_jax))
    print("mag diff max Mulens - microJAX: ", jnp.max(diff_mulens_jax))

    trajectory_mulens = my_model.get_trajectory(times= t + 2450000.0)
    x_vbbl, y_vbbl = -jnp.array(x), -jnp.array(y)

    diff_vbbl_jax_xy = jnp.array([x_vbbl, y_vbbl]) - jnp.array([tm, um])
    diff_mulens_jax_xy = jnp.array([trajectory_mulens.x, trajectory_mulens.y]) - jnp.array([tm, um])
    print("diff pos max VBM - MicroJAX:    ", jnp.max(jnp.sqrt(diff_vbbl_jax_xy[0]**2 + diff_vbbl_jax_xy[1]**2)))
    print("diff pos max Mulens - microJAX: ", jnp.max(jnp.sqrt(diff_mulens_jax_xy[0]**2 + diff_mulens_jax_xy[1]**2)))
    exit(1)



"""Adaptive lightcurve computation mixing multipole and inverse-ray methods.

This module computes microlensing lightcurves by defaulting to the fast
hexadecapole approximation and selectively invoking the full inverse-ray
finite-source integration near caustics or when required by accuracy tests.
"""

from functools import partial

import jax
import jax.numpy as jnp
from jax import jit, lax, vmap 

from microjax.inverse_ray.extended_source import mag_uniform, mag_limb_dark
from microjax.point_source import _images_point_source
from microjax.multipole import _mag_hexadecapole
from microjax.utils import *
from microjax.inverse_ray.cond_extended import _caustics_proximity_test, _planetary_caustic_test
from microjax.inverse_ray.cond_extended import test_full
from typing import Tuple

# Consistent array alias used across modules
Array = jnp.ndarray

@partial(jit,static_argnames=("r_resolution", "th_resolution", "u1", "delta_c", 
                              "bins_r", "bins_th", "margin_r", "margin_th", 
                              "Nlimb", "MAX_FULL_CALLS", "chunk_size"))
def mag_binary(
    w_points: Array,
    rho: float,
    r_resolution: int = 1000,
    th_resolution: int = 1000,
    u1: float = 0.0,
    delta_c: float = 0.01,
    Nlimb: int = 500,
    bins_r: int = 50,
    bins_th: int = 120,
    margin_r: float = 1.0,
    margin_th: float = 1.0,
    MAX_FULL_CALLS: int = 500,
    chunk_size: int = 100,
    **params,
) -> Array:
    """Binary-lens lightcurve with adaptive full-solve selection.

    Parameters follow the underlying inverse-ray integrators. The selection
    uses a proximity test to caustics plus planetary-caustic checks for small
    mass ratios. Returns magnifications per input time/position.
    """
    s = params.get("s", None)
    q = params.get("q", None)
    if s is None or q is None:
        raise ValueError("For nlenses=2, 's' and 'q' must be provided.") 
    
    nlenses = 2
    a = 0.5 * s
    e1 = q / (1.0 + q)
    _params = {**params, "a": a, "e1": e1}
    x_cm = a * (1.0 - q) / (1.0 + q)
    w_points_shifted = w_points - x_cm

    # test whether inverse-ray shooting needed or not. test==False means needed.
    z, z_mask = _images_point_source(w_points_shifted, nlenses=nlenses, a=a, e1=e1)
    mu_multi, delta_mu_multi = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params)
    test1 = _caustics_proximity_test(w_points_shifted, z, z_mask, rho, delta_mu_multi, nlenses=nlenses,  **_params)
    test2 = _planetary_caustic_test(w_points_shifted, rho, **_params)
    test = jnp.where(q < 0.01, test1 & test2, test1)

    if u1 == 0.0:
        def _mag_full(w):
            return mag_uniform(w, rho, nlenses = nlenses, r_resolution = r_resolution, th_resolution = th_resolution,
                               bins_r = bins_r, bins_th = bins_th, margin_r = margin_r, margin_th = margin_th, 
                               Nlimb = Nlimb, **_params)
    else:
        def _mag_full(w):
            return mag_limb_dark(w, rho, nlenses = nlenses, r_resolution = r_resolution, th_resolution= th_resolution,
                                 u1 = u1, delta_c = delta_c, bins_r = bins_r, bins_th = bins_th, margin_r = margin_r,
                                 margin_th = margin_th, Nlimb = Nlimb, **_params)
    
    idx_sorted = jnp.argsort(test)
    idx_full = idx_sorted[:MAX_FULL_CALLS]

    def chunked_vmap(func, data, chunk_size):
        N = data.shape[0]
        pad_len = (-N) % chunk_size
        chunks = jnp.pad(data, [(0, pad_len)] + [(0, 0)] * (data.ndim - 1)).reshape(-1, chunk_size, *data.shape[1:])
        return lax.map(lambda c: vmap(func)(c), chunks).reshape(-1, *data.shape[2:])[:N]
    
    def chunked_vmap_scan(func, data, chunk_size):
        N = data.shape[0]
        pad_len  = (-N) % chunk_size
        chunks   = jnp.pad(data, [(0, pad_len)] + [(0, 0)] * (data.ndim - 1)).reshape(-1, chunk_size, *data.shape[1:]) 
        #@jax.checkpoint
        def body(carry, chunk):
            #out = vmap(jax.checkpoint(func))(chunk)             
            out = vmap(func)(chunk)                            
            return carry, out               
        _, outs = lax.scan(body, None, chunks)   # outs → (T, chunk_size, ...)
        return outs.reshape(-1, *data.shape[2:])[:N]

    _mag_full = jax.checkpoint(_mag_full, policy=jax.checkpoint_policies.nothing_saveable, prevent_cse=False)
    mag_extended = chunked_vmap(_mag_full, w_points[idx_full], chunk_size)
    #mag_extended = chunked_vmap_scan(_mag_full, w_points[idx_full], chunk_size)
    mags = mu_multi.at[idx_full].set(mag_extended)
    mags = jnp.where(test, mu_multi, mags)
    return mags 

@partial(jit,static_argnames=("r_resolution", "th_resolution", "u1", "delta_c",
                              "bins_r", "bins_th", "margin_r", "margin_th", 
                              "Nlimb", "MAX_FULL_CALLS", "chunk_size"))
def mag_triple(
    w_points: Array,
    rho: float,
    r_resolution: int = 1000,
    th_resolution: int = 1000,
    u1: float = 0.0,
    delta_c: float = 0.01,
    Nlimb: int = 500,
    bins_r: int = 50,
    bins_th: int = 120,
    margin_r: float = 1.0,
    margin_th: float = 1.0,
    MAX_FULL_CALLS: int = 500,
    chunk_size: int = 50,
    **params,
) -> Array:
    """Triple-lens lightcurve with adaptive full-solve selection."""
    nlenses = 3
    s, q, q3 = params["s"], params["q"], params["q3"]
    a = 0.5 * s
    e1 = q / (1.0 + q + q3) 
    e2 = 1.0/(1.0 + q + q3)
    #r3 = r3 * jnp.exp(1j * psi)
    #_params = {"a": a, "r3": r3, "e1": e1, "e2": e2}
    _params = {**params, "a": a, "e1": e1, "e2": e2}
    #_params = {"a": a, "r3": r3, "e1": e1, "e2": e2, "q": q, "s": s, "q3": q3, "psi": psi}
    x_cm = a * (1.0 - q) / (1.0 + q)
    w_points_shifted = w_points - x_cm
    
    z, z_mask = _images_point_source(w_points_shifted, nlenses=nlenses, **_params) 
    mu_multi, delta_mu_multi = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params)
    test = jnp.zeros_like(w_points).astype(jnp.bool_)

    if u1 == 0.0:
        def _mag_full(w):
            return mag_uniform(w, rho, nlenses = nlenses, r_resolution = r_resolution, th_resolution = th_resolution,
                               bins_r = bins_r, bins_th = bins_th, margin_r = margin_r, margin_th = margin_th, 
                               Nlimb = Nlimb, **_params)
    else:
        def _mag_full(w):
            return mag_limb_dark(w, rho, nlenses = nlenses, r_resolution = r_resolution, th_resolution= th_resolution,
                                 u1 = u1, delta_c = delta_c, bins_r = bins_r, bins_th = bins_th, margin_r = margin_r,
                                 margin_th = margin_th, Nlimb = Nlimb, **_params)

    idx_sorted = jnp.argsort(test)
    idx_full = idx_sorted[:MAX_FULL_CALLS]

    def chunked_vmap(func, data, chunk_size):
        N = data.shape[0]
        pad_len = (-N) % chunk_size
        chunks = jnp.pad(data, [(0, pad_len)] + [(0, 0)] * (data.ndim - 1)).reshape(-1, chunk_size, *data.shape[1:])
        return lax.map(lambda c: vmap(func)(c), chunks).reshape(-1, *data.shape[2:])[:N]

    _mag_full = jax.checkpoint(_mag_full)
    mag_extended = chunked_vmap(_mag_full, w_points[idx_full], chunk_size)
    mags = mu_multi.at[idx_full].set(mag_extended)
    mags = jnp.where(test, mu_multi, mags)
    return mags 
    

if __name__ == "__main__":
    import time
    import jax
    jax.config.update("jax_enable_x64", True)
    #jax.config.update("jax_debug_nans", True)
    
    if(1):
        t0, u0, tE = 6.83640951e+03, 2.24211333e-01, 1.33559958e+02 
        s, q, alpha = 9.16157288e-01, 5.87559438e-04, jnp.deg2rad(1.00066409e+02)
        rho, pi_EN, pi_EE = 2.44003713e-03, 1.82341182e-01,9.58542572e-02

    if(0):
        q = 0.01
        s = 1.0
        alpha = jnp.deg2rad(10) 
        tE = 30 
        t0 = 0.0 
        u0 = 0.0 
        rho = 0.02

    nlenses = 2
    a = 0.5 * s
    e1 = q / (1.0 + q)
    _params = {"a": a, "e1": e1}
    x_cm = a * (1 - q) / (1 + q)

    num_points = 2000
    t  =  jnp.linspace(-1.0*tE + t0, 1.0*tE + t0, num_points)
    #t  =  jnp.linspace(-0.8*tE, 0.8*tE, num_points)
    tau = (t - t0)/tE
    y1 = -u0*jnp.sin(alpha) + tau*jnp.cos(alpha)
    y2 = u0*jnp.cos(alpha) + tau*jnp.sin(alpha) 
    w_points = jnp.array(y1 + y2 * 1j, dtype=complex)
    test_params = {"q": q, "s": s}  # Lens parameters

    Nlimb = 500
    r_resolution  = 1000
    th_resolution = 1000
    MAX_FULL_CALLS = 100

    cubic = True
    bins_r = 50
    bins_th = 120
    margin_r = 1.0
    margin_th= 1.0

    import MulensModel as mm
    import VBBinaryLensing
    VBBL = VBBinaryLensing.VBBinaryLensing()
    VBBL.a1 = 0.0
    VBBL.RelTol = 1e-5
    params_VBBL = [jnp.log(s), jnp.log(q), u0, alpha - jnp.pi, jnp.log(rho), jnp.log(tE), t0]

    #magn2  = lambda w0: jnp.array([mag_vbbl(w, rho) for w in w0])
    
    print("number of data points: %d"%(num_points))
    from microjax.point_source import mag_point_source, critical_and_caustic_curves
    mag_point_source(w_points, s=s, q=q)
    start = time.time()
    mags_poi = mag_point_source(w_points, s=s, q=q)
    mags_poi.block_until_ready()
    end = time.time()
    print("computation time: %.3f sec (%.3f ms) per points for point-source in microjax"%(end-start, 1000*(end - start)/num_points)) 

    from microjax.multipole import _mag_hexadecapole
    z, z_mask = _images_point_source(w_points - x_cm, nlenses=nlenses, **_params)
    _, _ = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params) 
    start = time.time()
    z, z_mask = _images_point_source(w_points - x_cm, nlenses=nlenses, **_params)
    mu_multi, delta_mu_multi = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params)
    mu_multi.block_until_ready()
    end = time.time()
    print("computation time: %.3f sec (%.3f ms per points) for hexadecapole in microjax"%(end-start, 1000*(end - start)/num_points)) 


    start = time.time()
    magnifications2, y1, y2 = jnp.array(VBBL.BinaryLightCurve(params_VBBL, t))
    magnifications2.block_until_ready() 
    end = time.time()
    print("computation time: %.3f sec (%.3f ms per points) for VBBinaryLensing"%(end - start,1000*(end - start)/num_points))

    _ = mag_binary(w_points, rho, s=s, q=q, r_resolution=r_resolution, th_resolution=th_resolution, cubic=cubic, 
                       Nlimb=Nlimb, bins_r=bins_r, bins_th=bins_th, margin_r=margin_r, margin_th=margin_th, MAX_FULL_CALLS=MAX_FULL_CALLS)
    #_ = mag_lc_uniform(w_points, rho, s=s, q=q, r_resolution=r_resolution, th_resolution=th_resolution, cubic=cubic, 
    #                   Nlimb=Nlimb, bins_r=bins_r, bins_th=bins_th, margin_r=margin_r, margin_th=margin_th, MAX_FULL_CALLS=MAX_FULL_CALLS)
    print("start computation with mag_lc_uniform, %d full calculation, %d rbin, %d thbin"%(MAX_FULL_CALLS, r_resolution, th_resolution))
    start = time.time()
    magnifications = mag_binary(w_points, rho, s=s, q=q, r_resolution=r_resolution, th_resolution=th_resolution,
                                    cubic=cubic, Nlimb=Nlimb, bins_r=bins_r, bins_th=bins_th, 
                                    margin_r=margin_r, margin_th=margin_th, MAX_FULL_CALLS=MAX_FULL_CALLS)
    end = time.time()
    print("computation time: %.3f sec (%.3f ms per points) for mag_lc_uniform in microjax"%(end-start, 1000*(end - start)/num_points))
   
    # Print out the result
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    import matplotlib.ticker as ticker
    import seaborn as sns
    sns.set_theme(style="ticks")

    critical_curves, caustic_curves = critical_and_caustic_curves(nlenses=2, npts=100, s=s, q=q)

    fig, ax_ = plt.subplots(2,1,figsize=(8,6), sharex=True, gridspec_kw=dict(hspace=0.1, height_ratios=[4,1]))
    ax  = ax_[0]
    ax1 = ax_[1]
    ax_in = inset_axes(ax,
        width="60%", height="60%", 
        bbox_transform=ax.transAxes,
        bbox_to_anchor=(0.35, 0.35, 0.6, 0.6)
    )
    ax_in.set_aspect(1)
    ax_in.set(xlabel="$\mathrm{Re}(w)$", ylabel="$\mathrm{Im}(w)$")
    for cc in caustic_curves:
        ax_in.plot(cc.real, cc.imag, color='red', lw=0.7)
    circles = [plt.Circle((xi,yi), radius=rho, fill=False, facecolor=None, ec="blue", zorder=2) 
               for xi, yi in zip(w_points.real, w_points.imag)
               ]
    c = mpl.collections.PatchCollection(circles, match_original=True, alpha=0.5)
    ax_in.add_collection(c)
    ax_in.set_aspect(1)
    ax_in.set(xlim=(-1., 1.2), ylim=(-1.0, 1.))
    ax_in.plot(-q/(1+q) * s, 0 , ".",c="k")
    ax_in.plot((1.0)/(1+q) * s, 0 ,".",c="k")

    ax.plot(t, magnifications, ".", label="microjax", zorder=1)
    ax.plot(t, magnifications2, "-", label="VBBinaryLensing", zorder=2)
    ylim = ax.get_ylim()
    #ax.plot(t, mags_poi, "--", label="point-source", zorder=-1, color="gray")
    ax.set_title("mag_lc_uniform")
    ax.grid(ls=":")
    ax.set_ylabel("magnification")
    ax.set_ylim(ylim[0], ylim[1])
    ax1.plot(t, jnp.abs(magnifications - magnifications2)/magnifications2, "-", ms=1)
    ax1.grid(ls=":")
    ax1.set_yticks(10**jnp.arange(-4, -2, 1))
    ax1.set_ylabel("relative diff")
    ax1.set_yscale("log")
    ax1.yaxis.set_major_locator(ticker.LogLocator(base=10.0, subs=[1.0, 10**-2, 10**-4, 10**-6], numticks=10))
    ax1.set_ylim(1e-6, 1e-2)
    ax.legend(loc="upper left")
    ax1.set_xlabel("time (days)")
    
    mu_multi, delta_mu_multi = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params)
    test1 = _caustics_proximity_test(
        w_points - x_cm, z, z_mask, rho, delta_mu_multi, nlenses=nlenses,  **_params 
    )
    test2 = _planetary_caustic_test(w_points - x_cm, rho, **_params)

    test = lax.cond(
        q < 0.01, 
        lambda:test1 & test2,
        lambda:test1,
    )
    ax.plot(t[~test], magnifications[~test], ".", color="red", zorder=20)
    print("full num: %d"%jnp.sum(~test))
    plt.show()
    #plt.savefig("z_fig/mag_lc.png", bbox_inches="tight", dpi=300)
    #print("z_fig/mag_lc.png")
    plt.close()



"""Conditions to decide when full inverse-ray integration is needed.

This module provides tests that determine whether the hexadecapole approximation
is sufficient for a given source position or whether to fall back to the full
inverse-ray integration. The logic is designed for binary and triple lenses.
"""

from functools import partial

import jax.numpy as jnp
from jax import jit, lax, vmap 
from microjax.point_source import _images_point_source
from microjax.multipole import _mag_hexadecapole
from typing import Tuple

# Consistent array alias used across inverse_ray and trajectory modules
Array = jnp.ndarray

@partial(jit, static_argnames=("nlenses"))
def test_full(w_points_shifted: Array, rho: float, nlenses: int = 2, **_params) -> Array:
    """Return a boolean mask where hexadecapole is sufficient (no full solve).

    For nlenses=2, combines a caustic-proximity test and a planetary-caustic
    exclusion. For other ``nlenses`` values it returns ``False`` everywhere.
    """
    e1 = _params["e1"]
    q = e1 / (1.0 - e1)
    if nlenses==2:
        # Compute hexadecapole approximation at every point and a test where it is sufficient
        z, z_mask = _images_point_source(w_points_shifted, nlenses=nlenses, **_params)
        mu_multi, delta_mu_multi = _mag_hexadecapole(z, z_mask, rho, nlenses=nlenses, **_params)
        test1 = _caustics_proximity_test(w_points_shifted, z, z_mask, rho, delta_mu_multi, nlenses=nlenses,  **_params)
        test2 = _planetary_caustic_test(w_points_shifted, rho, **_params)
        test = jnp.where(q < 0.01, test1 & test2, test1)
    else:
        test = jnp.zeros_like(w_points_shifted).astype(jnp.bool_)
    # test==True means no needs for inverse-ray shooting
    return test

@partial(jit, static_argnames=("nlenses"))
def _caustics_proximity_test(
    w: Array,
    z: Array,
    z_mask: Array,
    rho: float,
    delta_mu_multi: Array,
    nlenses: int = 2,
    c_m: float = 1e-02,
    gamma: float = 0.02,
    c_f: float = 4.0,
    rho_min: float = 1e-03,
    **params,
) -> Array:
    """Multipole accuracy and cusp proximity tests near caustics.

    Combines a magnitude threshold on the hexadecapole correction and a cusp
    proximity metric. Also checks for false images and filters them with a
    scale set by ``c_f``.
    """
    if nlenses == 2:
        a, e1 = params["a"], params["e1"]
        e2 = 1.0 - e1
        # Derivatives
        f = lambda z: - e1 / (z - a) - e2 / (z + a)
        f_p = lambda z: e1 / (z - a) ** 2 + e2 / (z + a) ** 2
        f_pp = lambda z: 2 * (e1 / (a - z) ** 3 - e2 / (a + z) ** 3)

    elif nlenses == 3:
        a, r3, e1, e2 = params["a"], params["r3"], params["e1"], params["e2"]
        # Derivatives
        f = lambda z: - e1 / (z - a) - e2 / (z + a) - (1 - e1 - e2) / (z + r3)
        f_p = (
            lambda z: e1 / (z - a) ** 2
            + e2 / (z + a) ** 2
            + (1 - e1 - e2) / (z + r3) ** 2
        )
        f_pp = (
            lambda z: 2 * (e1 / (a - z) ** 3 - e2 / (a + z) ** 3)
            + (1 - e1 - e2) / (z + r3) ** 3
        )
    zbar = jnp.conjugate(z)
    zhat = jnp.conjugate(w) - f(z)

    # Derivatives
    fp_z     = f_p(z)
    fpp_z    = f_pp(z)
    fp_zbar  = f_p(zbar)
    fp_zhat  = f_p(zhat)
    fpp_zbar = f_pp(zbar)
    J        = 1.0 - jnp.abs(fp_z * fp_zbar)

    # Multipole test and cusp test
    mu_cusp = 6 * jnp.imag(3 * fp_zbar**3.0 * fpp_z**2.0) / J**5 * (rho + rho_min)**2
    mu_cusp = jnp.sum(jnp.abs(mu_cusp) * z_mask, axis=0)
    test_multipole_and_cusp = gamma * mu_cusp + delta_mu_multi < c_m

    # False images test
    Jhat = 1 - jnp.abs(fp_z * fp_zhat)
    factor = jnp.abs(J * Jhat**2 / 
                     (Jhat*fpp_zbar*fp_z - jnp.conjugate(Jhat)  * fpp_z * fp_zbar * fp_zhat)
                     )
    test_false_images = 0.5 * (~z_mask * factor).sum(axis=0) > c_f * (rho + rho_min)
    test_false_images = jnp.where((~z_mask).sum(axis=0)==0, 
                                  jnp.ones_like(test_false_images, dtype=jnp.bool_), 
                                  test_false_images
                                  )
    return test_false_images & test_multipole_and_cusp

def _planetary_caustic_test(w: Array, rho: float, c_p: float = 2.0, **params) -> Array:
    """Exclude regions too close to planetary caustics for small mass ratios."""
    e1, a = params["e1"], params["a"]
    s = 2 * a
    q = e1 / (1.0 - e1)
    x_cm = (2*e1 - 1)*a
    w_pc = -1/s 
    delta_pc = 3*jnp.sqrt(q)/s
    return (w_pc - w).real**2 + (w_pc - w).imag**2 > c_p*(rho**2 + delta_pc**2)



import numpy as np

from microjax.fastlens.mag_fft_jax import magnification_disk as jx_magnification_disk
from microjax.fastlens.mag_fft_jax import A_point as jx_A_point


def test_jax_magnification_disk_A0_large_rho_matches():
    rho = 1e-2
    mag = jx_magnification_disk()
    a0_expected = np.sqrt(rho**2 + 4) / rho
    a = np.array(mag.A(0.0, rho))
    assert np.allclose(a, a0_expected, rtol=1e-10, atol=1e-12)


def test_jax_magnification_disk_small_rho_limits():
    rho = 1e-6
    mag = jx_magnification_disk()
    u = 100 * rho
    a = np.array(mag.A(u, rho))
    assert np.allclose(a, np.array(jx_A_point(u)), rtol=1e-6, atol=1e-8)


def test_jax_A_point_monotone_to_one():
    u = np.array([10.0, 30.0])
    ap = np.array(jx_A_point(u))
    assert np.all(ap > 1.0)
    assert (ap[1] - 1.0) < (ap[0] - 1.0)




"""Root finder utilities based on the Ehrlich–Aberth method.

This module implements a JAX-friendly version of the Ehrlich–Aberth (EA)
iteration to find all complex roots of a polynomial given in `polyval` order.

Design highlights
- Iteration uses `lax.while_loop` for early stopping by tolerance.
- Gradients are provided via implicit differentiation using `custom_jvp`,
  making reverse/forward AD robust regardless of the number of iterations.
- The derivative polynomial is precomputed from coefficients to avoid
  constructing it at every step; no reliance on `jnp.polyder`.

Notes on numerical behavior
- For nearly real coefficients, tiny imaginary parts in the output roots are
  zeroed (threshold ~ 10*tol) purely for presentation; the solver remains fully
  complex and differentiable.
- For ill-conditioned cases (e.g., near-multiple roots), the method can slow
  down; the implementation falls back to a Newton step when the EA denominator
  gets too small to improve stability.
"""

from typing import Optional, Tuple

import jax
import jax.numpy as jnp
from jax import lax, jit
from functools import partial
from jax import custom_jvp

# Default iteration cap for EA method; used if not specified.
_DEFAULT_MAX_ITER = 100

__all__ = [
    "poly_roots",
    "poly_roots_EA",
    "poly_roots_EA_multi",
]


def _as_complex(x: jax.Array) -> jax.Array:
    """Ensure complex dtype so real inputs can hold complex roots.

    Parameters
    - x: array-like; any real/complex dtype.

    Returns
    - Complex-typed array with a precision matching the input.
    """
    if jnp.iscomplexobj(x):
        return x
    # Promote to complex matching float precision
    dtype = jnp.result_type(x, jnp.complex64)
    return x.astype(dtype)


def _polyder_coeffs(coeffs: jax.Array) -> jax.Array:
    """Return derivative coefficients in `polyval` convention.

    If `coeffs = [c0, c1, ..., cn]` represents `c0*x^n + ... + cn`, then
    `d/dx coeffs = [c0*n, c1*(n-1), ..., c_{n-1}*1]`.

    For constant polynomials (degree 0), returns a zero array broadcast to the
    input shape.
    """
    n = coeffs.shape[-1] - 1
    if n <= 0:
        return jnp.zeros_like(coeffs)
    powers = jnp.arange(n, 0, -1, dtype=coeffs.real.dtype)
    return coeffs[..., :-1] * powers


def _cauchy_radius(coeffs: jax.Array) -> jax.Array:
    """Cauchy root bound radius using the leading coefficient.

    Parameters
    - coeffs: 1D array `[c0, c1, ..., cn]` in `polyval` order.

    Returns
    - A non-negative scalar array giving the radius R s.t. all roots lie in
      `|z| <= R` (simple bound `1 + max |c_i/c0|`).
    """
    c0 = coeffs[..., 0]
    eps = jnp.finfo(coeffs.real.dtype).eps
    denom = jnp.maximum(jnp.abs(c0), eps)
    # 1 + max |c_i / c0|
    r = 1.0 + jnp.max(jnp.abs(coeffs[..., 1:]) / denom[..., None], axis=-1)
    return r


def _ea_step(
    roots: jax.Array,
    coeffs: jax.Array,
    dcoeffs: jax.Array,
    eps: float,
) -> Tuple[jax.Array, jax.Array]:
    """Perform one Ehrlich–Aberth iteration step.

    Parameters
    - roots: shape `(n,)`, current root estimates.
    - coeffs: shape `(n+1,)`, polynomial coefficients (polyval order).
    - dcoeffs: shape `(n,)`, derivative coefficients (polyval order).
    - eps: small positive threshold to guard tiny denominators.

    Returns
    - new_roots: shape `(n,)`, updated root estimates.
    - update: shape `(n,)`, the applied update (for convergence checks).
    """
    # Evaluate polynomial and derivative at current roots
    f = jnp.polyval(coeffs, roots)
    fp = jnp.polyval(dcoeffs, roots)
    # Pairwise differences and reciprocal sum
    diffs = roots[:, None] - roots[None, :]
    inv_diffs = jnp.where(diffs == 0, 0.0 + 0.0j, 1.0 / diffs)
    s = jnp.sum(inv_diffs, axis=1)
    # Denominator stabilization: if nearly zero, fall back to Newton step
    den = fp - f * s
    den = jnp.where(jnp.abs(den) < eps, fp, den)
    update = f / den
    new_roots = roots - update
    return new_roots, update


@custom_jvp
def poly_roots_EA(
    coeffs: jax.Array,
    initial_roots: Optional[jax.Array] = None,
    tol: float = 1e-12,
    max_iter: int = _DEFAULT_MAX_ITER,
) -> jax.Array:
    """Compute all roots via the Ehrlich–Aberth method with early stopping.

    Parameters
    - coeffs: shape `(n+1,)`, polynomial coefficients `[c0, ..., cn]` with
      `polyval` order (i.e., `c0*x^n + ... + cn`). Real or complex.
    - initial_roots: optional, shape `(n,)` complex initial guesses. If not
      provided, points are placed uniformly on the circle given by a Cauchy
      bound.
    - tol: absolute update tolerance per root. A root is considered converged
      when `|Δ| <= tol`.
    - max_iter: maximum number of iterations.

    Returns
    - roots: shape `(n,)`, complex roots in unspecified order.

    AD behavior
    - Reverse/forward AD is supported via implicit differentiation (`custom_jvp`).
      The gradients do not depend on the number of EA iterations.
    """
    coeffs = _as_complex(coeffs)
    n = coeffs.shape[0] - 1
    assert n >= 1, "Polynomial degree must be >= 1"

    # Initial roots by Cauchy bound on a circle
    if initial_roots is None:
        r = _cauchy_radius(coeffs)
        k = jnp.arange(n)
        initial_roots = r * jnp.exp(2j * jnp.pi * k / n)
    else:
        initial_roots = _as_complex(initial_roots)

    dcoeffs = _polyder_coeffs(coeffs)
    eps = jnp.finfo(initial_roots.real.dtype).eps * 10

    def cond_fun(carry):
        it, roots, done = carry
        return jnp.logical_and(it < max_iter, jnp.any(~done))

    def body_fun(carry):
        it, roots, done = carry
        new_roots, update = _ea_step(roots, coeffs, dcoeffs, eps)
        err = jnp.abs(update)
        done_now = err <= tol
        return (it + 1, jnp.where(done[:, None], roots[:, None], new_roots[:, None]).squeeze(-1), done | done_now)

    init_done = jnp.zeros((n,), dtype=bool)
    init_state = (jnp.array(0, dtype=jnp.int32), initial_roots, init_done)
    _, roots, _ = lax.while_loop(cond_fun, body_fun, init_state)

    # If coefficients are (numerically) real, drop tiny imaginary parts in roots
    def _clean_real_roots(roots: jax.Array, coeffs: jax.Array) -> jax.Array:
        """Zero-out tiny imaginary parts for nearly real-coefficient polynomials."""
        imag = jnp.imag(coeffs)
        all_real = jnp.all(jnp.abs(imag) < 10 * tol)
        def clean(r):
            im = jnp.imag(r)
            re = jnp.real(r)
            im_clean = jnp.where(jnp.abs(im) < 10 * tol, 0.0, im)
            return re + 1j * im_clean
        return lax.cond(all_real, clean, lambda r: r, roots)

    roots = _clean_real_roots(roots, coeffs)
    return roots


@poly_roots_EA.defjvp
def _poly_roots_EA_jvp(primals, tangents):
    coeffs, initial_roots, tol, max_iter = primals
    dcoeffs, _, _, _ = tangents

    roots = poly_roots_EA(coeffs, initial_roots, tol, max_iter)
    # Implicit differentiation: f(roots; a) = 0
    dcoeffs = _as_complex(dcoeffs)
    n = coeffs.shape[0] - 1
    # derivative wrt z
    dcoeffs_poly = _polyder_coeffs(_as_complex(coeffs))
    df_dz = jnp.polyval(dcoeffs_poly, roots)
    # derivative wrt coefficients (Vandermonde order matches coeff order)
    exps = jnp.arange(n, -1, -1)  # integer exponents
    V = roots[:, None] ** exps[None, :]
    dz = - (V @ dcoeffs) / df_dz
    return roots, dz


def poly_roots_EA_multi(
    coeffs_matrix: jax.Array,
    custom_init: bool = False,
    initial_roots_matrix: Optional[jax.Array] = None,
    tol: float = 1e-12,
    max_iter: int = _DEFAULT_MAX_ITER,
) -> jax.Array:
    """Vectorized EA solver for batches of polynomials.

    Parameters
    - coeffs_matrix: array with shape `[..., n+1]`.
    - custom_init: whether to use `initial_roots_matrix` as initial guesses.
    - initial_roots_matrix: array with shape `[..., n]` if `custom_init=True`.
    - tol, max_iter: forwarded to `poly_roots_EA`.

    Returns
    - roots: array with shape `[..., n]`.
    """
    if custom_init:
        return jax.vmap(lambda c, r: poly_roots_EA(c, r, tol, max_iter))(coeffs_matrix, initial_roots_matrix)
    else:
        return jax.vmap(lambda c: poly_roots_EA(c, None, tol, max_iter))(coeffs_matrix)


@partial(jit, static_argnames=("custom_init",))
def poly_roots(
    coeffs: jax.Array,
    custom_init: bool = False,
    roots_init: Optional[jax.Array] = None,
) -> jax.Array:
    """Find all roots for a batch of polynomials using EA with early stop.

    Parameters
    - coeffs: array with shape `[..., n+1]` (polyval order). Real/complex.
    - custom_init: whether to use `roots_init` as initial guesses.
    - roots_init: array with shape `[..., n]` if `custom_init=True`.

    Returns
    - roots: array with shape `[..., n]` of complex roots in unspecified order.
    """
    ncoeffs = coeffs.shape[-1]
    output_shape = coeffs.shape[:-1] + (ncoeffs - 1,)
    coeffs_flat = coeffs.reshape((-1, ncoeffs))
    if custom_init:
        roots_init = roots_init.reshape((coeffs_flat.shape[0], ncoeffs - 1))
        roots = poly_roots_EA_multi(coeffs_flat, custom_init=True, initial_roots_matrix=roots_init)
    else:
        roots = poly_roots_EA_multi(coeffs_flat)
    return roots.reshape(output_shape)



import os
os.environ.setdefault("JAX_PLATFORMS", "cpu")

import numpy as np
import jax.numpy as jnp
import pytest
from tests.utils.gpu import gpu_tests_enabled, has_cuda, is_a100

from microjax.point_source import _images_point_source
from microjax.multipole import _mag_hexadecapole
from microjax.inverse_ray.lightcurve import mag_binary


def make_trajectory(u0, tE, t0, alpha, n=32, span=3.0):
    t = t0 + jnp.linspace(-span * tE, span * tE, n)
    tau = (t - t0) / tE
    y1 = -u0 * jnp.sin(alpha) + tau * jnp.cos(alpha)
    y2 = u0 * jnp.cos(alpha) + tau * jnp.sin(alpha)
    w_points = jnp.array(y1 + 1j * y2, dtype=complex)
    return t, w_points


def test_far_field_uses_multipole_matches_internal():
    # Choose far from caustics: large |w| and tiny q
    s, q, rho = 1.2, 1e-3, 1e-3
    a = 0.5 * s
    e1 = q / (1.0 + q)
    x_cm = a * (1.0 - q) / (1.0 + q)
    # Trajectory points far away
    _, w = make_trajectory(u0=5.0, tE=10.0, t0=0.0, alpha=jnp.deg2rad(30), n=16, span=1.0)
    w_shift = w - x_cm
    z, z_mask = _images_point_source(w_shift, nlenses=2, a=a, e1=e1)
    mu_multi, _ = _mag_hexadecapole(z, z_mask, rho, nlenses=2, a=a, e1=e1, s=s, q=q)

    mags = mag_binary(
        w,
        rho,
        s=s,
        q=q,
        r_resolution=64,
        th_resolution=64,
        Nlimb=64,
        bins_r=16,
        bins_th=32,
        margin_r=0.5,
        margin_th=0.5,
        MAX_FULL_CALLS=0,
        chunk_size=16,
    )
    # With MAX_FULL_CALLS=0, all points should use multipole path
    assert np.allclose(np.array(mags), np.array(mu_multi), rtol=1e-6, atol=1e-8)


@pytest.mark.parametrize(
    "s,q,u0,tE,rho,alpha",
    [
        (1.0, 1e-2, 0.05, 20.0, 5e-3, np.deg2rad(15.0)),
    ],
)
@pytest.mark.gpu
def test_binary_lightcurve_matches_vbbl(s, q, u0, tE, rho, alpha):
    if not (gpu_tests_enabled() and has_cuda() and is_a100()):
        pytest.skip("GPU/A100 not available or MICROJAX_GPU_TESTS not enabled")
    VB = pytest.importorskip("VBBinaryLensing")
    VBBL = VB.VBBinaryLensing()
    VBBL.a1 = 0.0
    VBBL.RelTol = 1e-5

    t0 = 0.0
    npts = 50
    t, w = make_trajectory(u0=u0, tE=tE, t0=t0, alpha=alpha, n=npts, span=2.0)
    params_vb = [jnp.log(s), jnp.log(q), u0, alpha - jnp.pi, jnp.log(rho), jnp.log(tE), t0]
    mag_vb, _, _ = jnp.array(VBBL.BinaryLightCurve(params_vb, t))

    mags = mag_binary(
        w,
        rho,
        s=s,
        q=q,
        r_resolution=128,
        th_resolution=128,
        Nlimb=128,
        bins_r=24,
        bins_th=48,
        margin_r=0.5,
        margin_th=0.5,
        MAX_FULL_CALLS=200,
        chunk_size=25,
    )

    diff = np.array(mags) - np.array(mag_vb)
    # Allow a modest tolerance due to discretization and algorithmic differences
    assert np.max(np.abs(diff)) < 2e-3



"""
python module for calculating microlensing magnification with finite source size effect
by Sunao Sugiyama
Jun 13, 2023
"""

import numpy as np
from microjax.fastlens import fftlog
from scipy.special import gamma
from scipy.special import j0, j1, jn, ellipk, ellipe

def A_point(u):
    return (u**2+2)/np.abs(u)/(u**2+4)**0.5

# FFT based magnification
class magnification:
    def __init__(self, fft_logumin=-6, fft_logumax=3, N_fft=1024, normalize_sk=True, rho_switch=1e-4, u_switch=10):
        """
        Args:
            fft_logumin      (int): minimum of FFT bin in u space (default=-6)
            fft_logumax      (int): maximum of FFT bin in u space (default=3)
            N_fft            (int): number of FFT bin in u space  (default=1024)
            normalize_sk    (bool): flag to normalize source profile in fourier space
        """
        # Defining FFT bin
        # Default choice is validated for rho > 1e-4 to ensure 0.3% precision
        self.fft_logumin  = fft_logumin 
        self.fft_logumax  = fft_logumax
        self.N_fft        = N_fft
        # flag to normalize source profile in fourier space
        self.normalize_sk = normalize_sk

        # The scale for rho and u to switch to use the approximate solution.
        # When rho < rho_switch, we use the approximate solution.
        # For u < u_switch*rho, we use Eq. (13) which is precomputed by init_small_rho below.
        # For u > u_switch*rho, we use point-source magnification.
        self.rho_switch = rho_switch
        self.u_switch   = u_switch

        # initialization
        self.init_Apk()
        self.init_Aext0()
        
    def init_Apk(self):
        """
        Initializing the FFT counter part of point-source magnification
        """
        u = np.logspace(self.fft_logumin, self.fft_logumax, self.N_fft)
        u2Au = ((u**2+2.0)/(u**2+4.0)**0.5/u - 1) * u**2
        h = fftlog.hankel(u, u2Au, nu=1.5, N_extrap_high=512, N_extrap_low=512)
        self.k, apk = h.hankel(0)
        self.apk = apk*2*np.pi

    def init_Aext0(self):
        """
        Implementation of A_ext0 in Eq. (13)
        """
        x    = np.logspace(-5, 5, 1024)
        dump = np.exp(-(x/100)**2)
        fx   = x * self.sk(x, 1) * dump
        h    = fftlog.hankel(x, fx, nu=1.5, N_pad=1024)
        u, aext0 = h.hankel(0)
        self.Aext0 = lambda x: np.interp(x, u, aext0)

    def sk(self, k, rho):
        # Implement Fourier counter part of source profile.
        pass

    def A0(self, rho):
        # Implement A(0, rho).
        pass

    def _A_for_small_rho(self, u, rho):
        """
        Evaluates the extended-source magnification for small rho.

        Aargs:
            u   (array)
            rho (float)
        """
        u = np.atleast_1d(np.abs(u))
        # Assign approximated solution: Eq. (13)
        a = np.ones(u.shape)
        idx = u < self.u_switch*rho
        x   = u[idx]/rho
        val = self.Aext0(x)/rho+1
        a[idx] = val
        # Assign approximated solution: point-source magnification
        idx = u >=self.u_switch*rho
        a[idx]= A_point(u[idx])
        return a

    def _A_for_large_rho(self, u, rho):
        """
        Evaluates the extended-source magnification for rho large enough

        Aargs:
            u   (array)
            rho (float)
        """
        u = np.atleast_1d(np.abs(u))
        # typical scale of source profile
        k_rho = 2*np.pi/rho
        # dumping factor to avoid noisy result
        dump = np.exp(-(self.k/k_rho/20)**2)
        # Fourier counter part of extended-source magnification
        cj = self.apk*self.k**2 * self.sk(self.k, rho) * dump
        # Hankel back the extended-source magnification
        h = fftlog.hankel(self.k, cj, nu=1.5, N_pad=512)
        u_fft, a_fft = h.hankel(0)
        a_fft = a_fft/2/np.pi
        a_fft = a_fft + 1
        
        # Truncate the result u>100 and append A(u=100)=1
        idx   = u_fft < 100
        u_fft = np.hstack([u_fft[idx], 100])
        a_fft = np.hstack([a_fft[idx], 1  ])

        # Assign values
        a     = np.ones(u.shape) * self.A0(rho)
        idx   = u>0
        a[idx]= np.interp(np.log10(u[idx]), np.log10(u_fft), a_fft, right=1)

        return a

    def A(self, u, rho):
        """
        Returns the extended-source magnification of microlensing light-curve.

        Args:
            u   (float): impact parameter in the lens plane normalized by Einstein angle.
            rho (float): source scale parameter.
        Returns:
            a   (float): magnification for extended-source profile.
        """
        u = np.atleast_1d(np.abs(u))
        inputs = (u, rho)

        if rho < self.rho_switch:
            return self._A_for_small_rho(u, rho)
        else:
            return self._A_for_large_rho(u, rho)

        
class magnification_disk(magnification):
    def sk(self, k, rho):
        """
        Implementation of Eq. (14)
        """
        k = np.atleast_1d(k)
        x   = k*rho
        a   = np.ones(x.shape)
        idx = x>0
        a[idx] = 2*j1(x[idx])/x[idx]
        return a

    def A0(self, rho):
        """
        Returns A_disk(0, rho)
        """
        return (rho**2+4)**0.5/rho

class magnification_limb(magnification):
    def __init__(self, n_limb, **kwargs):
        """
        This is for limb darkening profile, Eq. (6).
        
        Args:
            n_limb (int): order of limb darkening
        """
        self.n_limb = n_limb
        super().__init__(**kwargs)

    def sk(self, k, rho):
        """
        Implementation of Eq. (15)
        """
        k = np.atleast_1d(k)
        x   = k*rho
        nu  = 1+self.n_limb/2
        a   = np.ones(x.shape)*1.0/(self.n_limb+2)
        idx = x>0
        a[idx] = 2**nu*gamma(nu)*jn(nu, x[idx])/x[idx]**nu * nu
        return a

    def A0(self, rho):
        """
        Returns A_limb(0, rho)
        """
        if self.n_limb == 1:
            return (2+1) * (2*(rho**2+2)*ellipe(-rho**2/4)-(rho**2+4)*ellipk(-rho**2/4)) / 3.0/rho**3
        elif self.n_limb == 2:
            return (2+2) * (rho*(2+rho**2)*(4+rho**2)**0.5 - 8*np.arcsinh(rho/2)) / 4/rho**4
        elif self.n_limb >= 3:
            raise NotImplementedError

class magnification_log(magnification):
    """
    This is for logarithmic profile from https://doi.org/10.1086/110960.     
    """
    def su(self, u, rho):
        x   = u/rho
        ans = np.zeros(x.shape)
        # assign log profile
        idx = x<1
        sq = np.sqrt(1-x[idx]**2)
        ans[idx] = sq*np.log(sq)
        return x*np.log(x)
    
    def sk(self, k, rho):
        """
        Implementation of Fourier counterpart
        """
        pass

    def A0(self, rho):
        """
        Returns A_log(0, rho)
        """
        pass


import numpy as np
import pytest

from microjax.fastlens.fftlog import fftlog as np_fftlog
from microjax.fastlens.fftlog_jax import fftlog as jx_fftlog


def make_input():
    x = np.logspace(-2, 2, 256)
    fx = np.exp(-x**2)
    return x, fx


@pytest.mark.parametrize("ell", [0, 1, 2])
def test_fftlog_derivative_variants_numpy(ell):
    x, fx = make_input()
    fl = np_fftlog(x, fx, nu=1.5, N_extrap_low=0, N_extrap_high=0, N_pad=64)
    # Basic calls shouldn't raise and should yield finite arrays
    for method in (fl.fftlog_dj, fl.fftlog_ddj, fl.fftlog_jsqr):
        y, Fy = method(ell)
        assert y.shape == Fy.shape
        assert np.all(np.diff(y) > 0)
        assert np.all(np.isfinite(Fy))


@pytest.mark.parametrize("ell", [0, 1, 2])
def test_fftlog_derivative_variants_jax_vs_numpy_close(ell):
    x, fx = make_input()
    np_fl = np_fftlog(x, fx, nu=1.5, N_extrap_low=0, N_extrap_high=0, N_pad=64)
    jx_fl = jx_fftlog(x, fx, nu=1.5, N_extrap_low=0, N_extrap_high=0, N_pad=64)

    pairs = [
        (np_fl.fftlog_dj, jx_fl.fftlog_dj),
        (np_fl.fftlog_ddj, jx_fl.fftlog_ddj),
        (np_fl.fftlog_jsqr, jx_fl.fftlog_jsqr),
    ]
    for np_m, jx_m in pairs:
        y_np, Fy_np = np_m(ell)
        y_jx, Fy_jx = jx_m(ell)
        assert np.allclose(y_np, y_jx, rtol=1e-6, atol=1e-8)
        sl = slice(10, -10)
        assert np.allclose(Fy_np[sl], np.array(Fy_jx)[sl], rtol=5e-3, atol=5e-4)




import numpy as np
import pytest

from microjax.fastlens.fftlog import hankel as np_hankel
from microjax.fastlens.fftlog_jax import hankel as jax_hankel


def make_gaussian():
    # Log-spaced grid and Gaussian function
    x = np.logspace(-2, 2, 256)
    fx = np.exp(-x**2)
    return x, fx


@pytest.mark.parametrize("n", [0, 1])
def test_hankel_numpy_vs_jax_close(n):
    x, fx = make_gaussian()
    nu = 1.5
    # Prefer zero extrapolation and modest zero-padding to avoid edge NaNs
    np_h = np_hankel(x, fx, nu=nu, N_extrap_low=0, N_extrap_high=0, c_window_width=0.25, N_pad=64)
    jx_h = jax_hankel(x, fx, nu=nu, N_extrap_low=0, N_extrap_high=0, c_window_width=0.25, N_pad=64)

    y_np, Fy_np = np_h.hankel(n)
    y_jx, Fy_jx = jx_h.hankel(n)

    # Grids should match closely
    assert np.allclose(y_np, y_jx, rtol=1e-6, atol=1e-8)
    # Transforms should be close up to small numerical differences
    # Compare only on the central region to avoid extrap edge effects
    sl = slice(10, -10)
    assert np.allclose(Fy_np[sl], np.array(Fy_jx)[sl], rtol=5e-3, atol=5e-4)


def test_hankel_basic_properties():
    x = np.logspace(-2, 2, 256)
    fx = np.exp(-0.5 * x**2)
    h = np_hankel(x, fx, nu=1.5, N_extrap_low=0, N_extrap_high=0, N_pad=64)
    y, Fy = h.hankel(0)
    # Shapes align and y is increasing
    assert y.shape == Fy.shape
    assert np.all(np.diff(y) > 0)
    # Positive input should yield finite output
    assert np.all(np.isfinite(Fy))



"""Inverse-ray integration utilities for finite-source microlensing.

Submodules
- boundary: smoothed membership and boundary factors
- limb_darkening: limb-darkening intensity profiles
- merge_area: region construction for polar integration
- extended_source: core integrators (uniform, limb-darkened)
- cond_extended: tests to select between multipole and full solve
- lightcurve: adaptive lightcurve mixing multipole and full inverse-ray
"""

__all__ = [
    "boundary",
    "limb_darkening",
    "merge_area",
    "extended_source",
    "cond_extended",
    "lightcurve",
]
